# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import shutil
import sys
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.time import Time
from hydra.conf import HydraConf, JobConf
from hydra_zen import make_config, zen, ZenStore

# add parent directory of __file__ to sys.path, if isn't already included
if str(Path(__file__).parents[1]) not in sys.path:
    sys.path.append(str(Path(__file__).parents[1]))

from madlib import SensorCollection
from madlib._utils import MadlibException
from scripts import dropout
import pytest


PROJECT_ROOT = str(Path(__file__).parents[1])


def get_default_overrides() -> list:
    return [
        f"hydra.sweep.dir={PROJECT_ROOT}/tests/outputs/multirun/"
        + "${now:%Y-%m-%d}/${now:%H-%M-%S}",
        "hydra.sweep.subdir=${hydra.job.num}",
        "cloud_prob=1.0",
        "cloud_duration_mean=3600,7200",
        "cloud_duration_std=0.0",
    ]


def add_datetime(df: pd.DataFrame) -> None:
    """Adds a datetime64 column (inplace) to a DataFrame with an existing "MJD" column

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame with an existing "MJD" column
    """
    ts_mjd = Time(df["MJD"], format="mjd")
    ts_datetime = ts_mjd.to_value("datetime64")
    df["Datetime"] = ts_datetime
    df.sort_values("MJD", inplace=True)


class TestArgs:
    """Test behavior of argument parsing"""

    def test_argument_defaults(self):
        """Test default argument values"""
        inputs = [
            "--path",
            "inputs/complete.impulse.csv",
            "--sensor_yaml",
            "/path/to/sensor.yaml",
        ]

        parser = dropout.parseArgs()
        args = parser.parse_args(inputs)

        # Count number of input arguments
        assert len(args.__dict__) == 9

        assert isinstance(args.path, str)
        assert args.path == "inputs/complete.impulse.csv"
        assert isinstance(args.sensor_yaml, str)
        assert args.sensor_yaml == "/path/to/sensor.yaml"
        assert isinstance(args.cloud_prob, str)
        assert args.cloud_prob == "0.5"
        assert isinstance(args.cloud_duration_mean, str)
        assert args.cloud_duration_mean == "10800.0"
        assert isinstance(args.cloud_duration_std, str)
        assert args.cloud_duration_std == "3600.0"
        assert isinstance(args.num_runs, int)
        assert args.num_runs == 1
        assert isinstance(args.save_copy_of_original, bool)
        assert args.save_copy_of_original == False
        assert isinstance(args.save_plots, bool)
        assert args.save_plots == False
        assert isinstance(args.submitit, str)
        assert args.submitit == ""

    def test_argument_inputs(self):
        """Test that input values work as expected"""
        inputs = [
            "--path",
            "inputs/complete.impulse.csv",
            "--sensor_yaml",
            "/path/to/sensor.yaml",
            "--cloud_prob",
            "0.1",
            "--cloud_duration_mean",
            "60.0",
            "--cloud_duration_std",
            "10.0",
            "--num_runs",
            "10",
            "--save_copy_of_original",
            "--save_plots",
            "--submitit",
            PROJECT_ROOT + "/configs/submitit_slurm_template.json",
        ]

        parser = dropout.parseArgs()
        args = parser.parse_args(inputs)

        # Count number of input arguments
        assert len(args.__dict__) == 9

        assert isinstance(args.path, str)
        assert args.path == "inputs/complete.impulse.csv"
        assert isinstance(args.sensor_yaml, str)
        assert args.sensor_yaml == "/path/to/sensor.yaml"
        assert isinstance(args.cloud_prob, str)
        assert args.cloud_prob == "0.1"
        assert isinstance(args.cloud_duration_mean, str)
        assert args.cloud_duration_mean == "60.0"
        assert isinstance(args.cloud_duration_std, str)
        assert args.cloud_duration_std == "10.0"
        assert isinstance(args.num_runs, int)
        assert args.num_runs == 10
        assert isinstance(args.save_copy_of_original, bool)
        assert args.save_copy_of_original == True
        assert isinstance(args.save_plots, bool)
        assert args.save_plots == True
        assert isinstance(args.submitit, str)
        assert args.submitit == PROJECT_ROOT + "/configs/submitit_slurm_template.json"


class TestDropout:
    """Test dropout core behavior"""

    def test_dropout_general_functionality_impulse(self):
        """Test the core functionality of the dropout script with impulse maneuver datasets"""

        # customize Hydra's configuration
        store1 = ZenStore(overwrite_ok=True)
        store1(HydraConf(job=JobConf(chdir=True)))
        store1.add_to_hydra_store(overwrite_ok=True)

        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        DropoutConfig = make_config(
            path=PROJECT_ROOT + "/tests/inputs/complete.impulse.csv",
            sensor_list=sensor_list,
            sensor_idx=None,
            cloud_prob=0.5,
            cloud_duration_mean=10800.0,
            cloud_duration_std=3600.0,
            num_runs=1,
            save_copy_of_original=True,
            save_plots=True,
        )

        overrides = get_default_overrides()

        output_dir = dropout.launcher(
            config=DropoutConfig,
            task_function=task_function,
            overrides=overrides,
            submitit="",
            return_output_dir=True,
        )

        assert output_dir is not None
        # Check if data files exist and that data was "dropped"
        # Top level
        assert len(list(output_dir.glob("multirun.yaml"))) > 0
        assert len(list(output_dir.glob("*/"))) == 2
        # Sweep Param 1
        assert (output_dir / "0/output.csv").exists()
        df = pd.read_csv(output_dir / "0/output.csv")
        assert (output_dir / "0/output_dropped__0.csv").exists()
        df_drop = pd.read_csv(output_dir / "0/output_dropped__0.csv")
        assert df_drop.shape[0] < df.shape[0]
        assert len(list((output_dir / "0").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "0").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)
        assert (output_dir / "0/plots").exists()
        assert len(list((output_dir / "0/plots").rglob("*.png"))) > 0
        # Sweep Param 2
        assert (output_dir / "1/output.csv").exists()
        assert (output_dir / "1/output_dropped__0.csv").exists()
        df_drop = pd.read_csv(output_dir / "1/output_dropped__0.csv")
        assert df_drop.shape[0] < df.shape[0]
        assert len(list((output_dir / "1").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "1").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)
        assert (output_dir / "0/plots").exists()
        assert len(list((output_dir / "0/plots").rglob("*.png"))) > 0

        # Test plotting edge cases ---------------------------------------------|
        # Add Datetime column to DataFrames (required for plotting routines)
        add_datetime(df)
        add_datetime(df_drop)

        # Plot edge cases: sensor_ID -------------------------------------------|
        # save_without_displaying is False
        sensor_ID = "A1"
        assert (
            dropout.plot_sensor_ID(
                sensor_ID,
                df=df,
                df_drop=df_drop,
                sensor_list=sensor_list,
                save_without_displaying=False,
            )
            is None
        )

        # Plot sensor_ID does not exist
        sensor_ID = "INVALID_SENSOR_ID"
        assert (
            dropout.plot_sensor_ID(
                sensor_ID,
                df=df,
                df_drop=df_drop,
                sensor_list=sensor_list,
                save_without_displaying=False,
            )
            is None
        )

        # Plot edge cases: seq_ID ----------------------------------------------|
        # KeyError Exception Case
        seq_ID = 99999999
        assert (
            dropout.plot_seq_id(
                seq_ID,
                df=df,
                df_drop=df_drop,
                save_without_displaying=False,
            )
            is None
        )

        # save_without_displaying is False
        seq_ID = 0
        assert (
            dropout.plot_seq_id(
                seq_ID,
                df=df,
                df_drop=df_drop,
                save_without_displaying=False,
            )
            is None
        )

        # Remove multirun folder in entirety
        shutil.rmtree(PROJECT_ROOT + "/tests/outputs/multirun")

    def test_dropout_none_type_return(self):
        """Test dropout function with None type return"""

        # customize Hydra's configuration
        store1 = ZenStore(overwrite_ok=True)
        store1(HydraConf(job=JobConf(chdir=True)))
        store1.add_to_hydra_store(overwrite_ok=True)

        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        DropoutConfig = make_config(
            path=PROJECT_ROOT + "/tests/inputs/complete.impulse.csv",
            sensor_list=sensor_list,
            sensor_idx=None,
            cloud_prob=0.5,
            cloud_duration_mean=10800.0,
            cloud_duration_std=3600.0,
            num_runs=1,
            save_copy_of_original=True,
            save_plots=False,
        )

        overrides = get_default_overrides()

        output = dropout.launcher(
            config=DropoutConfig,
            task_function=task_function,
            overrides=overrides,
            submitit="",
            return_output_dir=False,
        )
        assert output is None

        # Remove multirun folder in entirety
        shutil.rmtree(PROJECT_ROOT + "/tests/outputs/multirun")

    def test_dropout_with_submitit_json_inputs(self):
        """Tests the dropout script while using JSON file paths for the submitit argument"""

        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        DropoutConfig = make_config(
            path=PROJECT_ROOT + "/tests/inputs/complete.impulse.csv",
            sensor_list=sensor_list,
            sensor_idx=None,
            cloud_prob=0.5,
            cloud_duration_mean=10800.0,
            cloud_duration_std=3600.0,
            num_runs=1,
            save_copy_of_original=True,
            save_plots=False,
        )

        # Test with submitit config
        try:
            # Try to use `srun` if on a system with SLURM
            overrides = get_default_overrides()

            output_dir = dropout.launcher(
                config=DropoutConfig,
                task_function=task_function,
                overrides=overrides,
                submitit=PROJECT_ROOT + "/configs/submitit_slurm_template.json",
                return_output_dir=True,
            )
            assert output_dir is not None
        except RuntimeError:
            # RuntimeError occurs if running these tests on a machine without SLURM,
            # use `submitit_test.json` which uses `hydra/launcher=submitit_local` in
            # this case.
            overrides = get_default_overrides()

            output_dir = dropout.launcher(
                config=DropoutConfig,
                task_function=task_function,
                overrides=overrides,
                submitit=PROJECT_ROOT + "/tests/inputs/submitit_test.json",
                return_output_dir=True,
            )
            assert output_dir is not None

        # Test with submitit invalid config
        with pytest.raises(MadlibException):
            overrides = get_default_overrides()

            dropout.launcher(
                config=DropoutConfig,
                task_function=task_function,
                overrides=overrides,
                submitit=PROJECT_ROOT + "/tests/inputs/submitit_invalid.json",
            )

        # Remove multirun folder in entirety
        shutil.rmtree(PROJECT_ROOT + "/tests/outputs/multirun")

    def test_dropout_general_functionality_continuous(self):
        """Test the core functionality of the dropout script with continuous maneuver datasets"""
        # customize Hydra's configuration
        store2 = ZenStore()
        store2(HydraConf(job=JobConf(chdir=True)))
        store2.add_to_hydra_store(overwrite_ok=True)

        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        DropoutConfig = make_config(
            path=PROJECT_ROOT + "/tests/inputs/complete.continuous.csv",
            sensor_list=sensor_list,
            sensor_idx=None,
            cloud_prob=0.5,
            cloud_duration_mean=10800.0,
            cloud_duration_std=3600.0,
            num_runs=1,
            save_copy_of_original=True,
            save_plots=True,
        )

        overrides = get_default_overrides()

        output_dir = dropout.launcher(
            config=DropoutConfig,
            task_function=task_function,
            overrides=overrides,
            submitit="",
            return_output_dir=True,
        )

        assert output_dir is not None
        # Check if data files exist and that data was "dropped"
        # Top level
        assert len(list(output_dir.glob("multirun.yaml"))) > 0
        assert len(list(output_dir.glob("*/"))) == 2
        # Sweep Param 1
        assert (output_dir / "0/output.csv").exists()
        df = pd.read_csv(output_dir / "0/output.csv")
        assert (output_dir / "0/output_dropped__0.csv").exists()
        df_dropped = pd.read_csv(output_dir / "0/output_dropped__0.csv")
        assert df_dropped.shape[0] < df.shape[0]
        assert len(list((output_dir / "0").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "0").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)
        # Sweep Param 2
        assert (output_dir / "1/output.csv").exists()
        assert (output_dir / "1/output_dropped__0.csv").exists()
        df_dropped = pd.read_csv(output_dir / "1/output_dropped__0.csv")
        assert df_dropped.shape[0] < df.shape[0]
        assert len(list((output_dir / "1").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "1").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)

        # Remove multirun folder in entirety
        shutil.rmtree(PROJECT_ROOT + "/tests/outputs/multirun")

    def test_dropout_fn_edge_cases(self):
        """Test dropout core functionality edge cases"""

        # Bad sensor list
        bad_sensor_list = [
            "this",
            "is",
            "a",
            "bad",
            "sensor_list",
        ]

        # Test bad sensor list (i.e. elements are not dicts)
        assert (
            dropout.dropout(
                path=PROJECT_ROOT + "/tests/inputs/complete.impulse.csv",
                sensor_list=bad_sensor_list,  # type: ignore
                sensor_idx=None,
                cloud_prob=0.5,
                cloud_duration_mean=10800.0,
                cloud_duration_std=3600.0,
                num_runs=1,
            )
            is None
        )

        # More than 1 num_run and save_plots = False
        # customize Hydra's configuration
        store4 = ZenStore(overwrite_ok=True)
        store4(HydraConf(job=JobConf(chdir=True)))
        store4.add_to_hydra_store(overwrite_ok=True)

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        DropoutConfig = make_config(
            path=PROJECT_ROOT + "/tests/inputs/complete.impulse.csv",
            sensor_list=sensor_list,
            sensor_idx=None,
            cloud_prob=0.5,
            cloud_duration_mean=10800.0,
            cloud_duration_std=3600.0,
            num_runs=2,
            save_copy_of_original=True,
            save_plots=False,
        )

        overrides = get_default_overrides()

        output_dir = dropout.launcher(
            config=DropoutConfig,
            task_function=task_function,
            overrides=overrides,
            submitit="",
            return_output_dir=True,
        )

        assert output_dir is not None
        # Check if data files exist and that data was "dropped"
        # Top level
        assert len(list(output_dir.glob("multirun.yaml"))) > 0
        assert len(list(output_dir.glob("*/"))) == 2
        # Sweep Param 1
        assert (output_dir / "0/output.csv").exists()
        df = pd.read_csv(output_dir / "0/output.csv")
        assert (output_dir / "0/output_dropped__0.csv").exists()
        df_drop = pd.read_csv(output_dir / "0/output_dropped__0.csv")
        assert df_drop.shape[0] < df.shape[0]
        assert len(list((output_dir / "0").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "0").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)

        # Remove multirun folder in entirety
        shutil.rmtree(PROJECT_ROOT + "/tests/outputs/multirun")

    def test_dropout_low_probability(self):
        """Test the core functionality of the dropout script when probability of a weather event is low (zero)"""
        # customize Hydra's configuration
        store4 = ZenStore(overwrite_ok=True)
        store4(HydraConf(job=JobConf(chdir=True)))
        store4.add_to_hydra_store(overwrite_ok=True)

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        # zen wrapper that will auto-extract, resolve, and instantiate fields from
        # an input config based on dropout()'s signature
        task_function = zen(dropout.dropout)

        DropoutConfig = make_config(
            path=PROJECT_ROOT + "/tests/inputs/complete.impulse.csv",
            sensor_list=sensor_list,
            sensor_idx=None,
            cloud_prob=0.5,
            cloud_duration_mean=10800.0,
            cloud_duration_std=3600.0,
            num_runs=1,
            save_copy_of_original=True,
            save_plots=True,
        )

        overrides = [
            f"hydra.sweep.dir={PROJECT_ROOT}/tests/outputs/multirun/"
            + "${now:%Y-%m-%d}/${now:%H-%M-%S}",
            "hydra.sweep.subdir=${hydra.job.num}",
            "cloud_prob=0.0",
            "cloud_duration_mean=3600,7200",
            "cloud_duration_std=0.0",
        ]

        output_dir = dropout.launcher(
            config=DropoutConfig,
            task_function=task_function,
            overrides=overrides,
            submitit="",
            return_output_dir=True,
        )

        assert output_dir is not None
        # Check if data files exist and that data was "dropped"
        # Top level
        assert len(list(output_dir.glob("multirun.yaml"))) > 0
        assert len(list(output_dir.glob("*/"))) == 2
        # Sweep Param 1
        assert (output_dir / "0/output.csv").exists()
        df = pd.read_csv(output_dir / "0/output.csv")
        assert (output_dir / "0/output_dropped__0.csv").exists()
        df_drop = pd.read_csv(output_dir / "0/output_dropped__0.csv")
        assert len(list((output_dir / "0").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "0").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)
        assert (output_dir / "0/plots").exists()
        assert len(list((output_dir / "0/plots").rglob("*.png"))) > 0
        # Sweep Param 2
        assert (output_dir / "1/output.csv").exists()
        assert (output_dir / "1/output_dropped__0.csv").exists()
        df_drop = pd.read_csv(output_dir / "1/output_dropped__0.csv")
        assert len(list((output_dir / "1").glob("*.pkl"))) > 0
        sensor_list = pd.read_pickle(list((output_dir / "1").glob("*.pkl"))[0])
        assert isinstance(sensor_list, list)
        assert (output_dir / "0/plots").exists()
        assert len(list((output_dir / "0/plots").rglob("*.png"))) > 0

        # Remove multirun folder in entirety
        shutil.rmtree(PROJECT_ROOT + "/tests/outputs/multirun")

    def test_dropout_get_sensor(self):
        """Test dropout.get_sensor()"""
        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        sensor = dropout.get_sensor(sensor_list, "INVALID_SENSOR_ID")
        assert sensor is None

    def test_dropout_update_sensor_weather_stats_with_idx(self):
        """Test dropout.update_sensor_weather_stats() index parameter"""
        # Parse the sensor YAML file
        sensor_data = SensorCollection.paramsFromYAML(
            PROJECT_ROOT + "/configs/sample_sensor_network.yaml"
        )
        sensor_list = [sensor_data[key] for key in sensor_data.keys()]

        idx = 1

        # check initial values
        sensor = sensor_list[idx]
        sensor_id_original = sensor["id"]
        assert np.isclose(sensor["weather"]["cloud_prob"], 0.5)
        assert np.isclose(sensor["weather"]["cloud_duration_mean"], 10800.0)
        assert np.isclose(sensor["weather"]["cloud_duration_std"], 7200.0)

        # update sensor using idx
        sensor_list = dropout.update_sensor_weather_stats(
            sensor_list,
            idx=idx,
            cloud_prob=0.555,
            cloud_duration_mean=12345.0,
            cloud_duration_std=321.0,
        )

        # check that the values were changed
        sensor = sensor_list[idx]
        assert np.isclose(sensor["weather"]["cloud_prob"], 0.555)
        assert np.isclose(sensor["weather"]["cloud_duration_mean"], 12345.0)
        assert np.isclose(sensor["weather"]["cloud_duration_std"], 321.0)
        assert sensor["id"] == sensor_id_original



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_propagate.py
Description: This file contains a unit tests which propagates a satellite,
    with and without a manuever, and compares the final lat&lon values to
    known values. This unit test in itself provides ~80% code coverage.
"""

import pathlib
import sys

import astroforge as af
import numpy as np
from astropy.time import Time

# add parent directory of __file__ to sys.path, if isn't already included
if str(pathlib.Path(__file__).parents[1]) not in sys.path:
    sys.path.append(str(pathlib.Path(__file__).parents[1]))
import madlib


def test_sat_impulse_maneuver(seed=4445):
    np.random.seed(seed)
    # --- orbital state
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- maneuver definition
    man_time = epoch + 5.0 / 24
    man_dv = np.array([0.0, 5.0, 0.0]) / 1000
    man = madlib.ImpulsiveManeuver(man_time, man_dv)

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)
    sat1 = madlib.Satellite(epoch, x0, v0, a0, man)

    # --- setup sensor
    SST_PARAMS = {
        "lat": -21.89567,
        "lon": 114.0898731,
        "alt": 0.067845,
        "dra": 0.3,
        "ddec": 0.3,
        "obs_per_collect": (3, 5),
        "obs_time_spacing": 1.5,
        "collect_gap_mean": 7200,
        "collect_gap_std": 800,
        # "obs_limits": {"el": (20.0, 91.0)},
        "obs_limits": None,
        "id": "SST",
    }

    sensor = madlib.GroundOpticalSensor(**SST_PARAMS)
    times = sensor.generate_obs_timing(epoch, epoch + 2)  # + 2 [hours]

    manuv_f = sensor.observe(sat0, times).pos_truth
    manuv_t = sensor.observe(sat1, times).pos_truth
    manuv_f_dict = {
        "ra": np.array([x.ra for x in manuv_f]),
        "dec": np.array([x.dec for x in manuv_f]),
    }
    manuv_t_dict = {
        "ra": np.array([x.ra for x in manuv_t]),
        "dec": np.array([x.dec for x in manuv_t]),
    }

    assert all(
        np.isclose(
            np.array(
                [
                    manuv_f_dict["ra"][-1],
                    manuv_f_dict["dec"][-1],
                    manuv_t_dict["ra"][-1],
                    manuv_t_dict["dec"][-1],
                ]
            ),
            np.array(
                [
                    28.88567036042452,
                    3.295301473939045,
                    25.393658788650338,
                    3.3090154095997817,
                ]
            ),
        )
    )


def test_sat_continuous_maneuver(seed=4445):
    np.random.seed(seed)
    # --- orbital state
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- maneuver definition
    # ContinuousThrust Function Definition
    def acc_f(t):
        return np.array([0.0, 1.0e-7, 0.0])

    acc_t_range = (epoch, epoch + 3)
    # ContinuousManeuver Definition
    man = madlib.ContinuousManeuver(acc_f, acc_t_range)

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)
    sat1 = madlib.ContinuousThrustSatellite(epoch, x0, v0, a0, man)

    # --- setup sensor
    SST_PARAMS = {
        "lat": -21.89567,
        "lon": 114.0898731,
        "alt": 0.067845,
        "dra": 0.3,
        "ddec": 0.3,
        "obs_per_collect": (3, 5),
        "obs_time_spacing": 1.5,
        "collect_gap_mean": 7200,
        "collect_gap_std": 800,
        # "obs_limits": {"el": (20.0, 91.0)},
        "obs_limits": None,
        "id": "SST",
    }

    sensor = madlib.GroundOpticalSensor(**SST_PARAMS)
    times = sensor.generate_obs_timing(epoch, epoch + 2)  # + 2 [hours]

    manuv_f = sensor.observe(sat0, times).pos_truth
    manuv_t = sensor.observe(sat1, times).pos_truth
    manuv_f_dict = {
        "ra": np.array([x.ra for x in manuv_f]),
        "dec": np.array([x.dec for x in manuv_f]),
    }
    manuv_t_dict = {
        "ra": np.array([x.ra for x in manuv_t]),
        "dec": np.array([x.dec for x in manuv_t]),
    }
    dt = np.array([Time(x.mjd, format="mjd").datetime for x in manuv_f])

    assert all(
        np.isclose(
            np.array(
                [
                    manuv_f_dict["ra"][-1],
                    manuv_f_dict["dec"][-1],
                    manuv_t_dict["ra"][-1],
                    manuv_t_dict["dec"][-1],
                ]
            ),
            np.array(
                [
                    28.88567036042452,
                    3.295301473939045,
                    30.024913619492622,
                    3.3128772755122093,
                ]
            ),
        )
    )



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_madlib_sensor.py
Description: This file contains unit tests which test "edge cases" in
    the madlib._sensor module.
"""

import astroforge as af
import numpy as np
from astropy.time import Time

import madlib
from madlib._utils import MadlibException
from madlib._sensor_collection import SensorCollection, SensorException

import pytest


seed = 4445

yaml = "configs/sample_sensor_network.yaml"


def test_SensorCollection_init_from_params():
    # Sensors
    sensor_params = SensorCollection.paramsFromYAML(yaml)
    sensor_objects = [
        madlib.GroundOpticalSensor(**params) for key, params in sensor_params.items()
    ]
    sensors = SensorCollection(sensor_objects)

    assert isinstance(sensors, madlib._sensor_collection.SensorCollection)
    assert sensors.sensorList == sensor_objects
    assert sensors.numSensors == len(sensor_objects)
    assert sensors.obsTimes == None


def test_SensorCollection_init_from_yaml():
    sensors = SensorCollection.fromYAML(yaml)

    assert isinstance(sensors, madlib._sensor_collection.SensorCollection)
    assert sensors.numSensors == 11
    assert sensors.obsTimes == None


def test_SensorCollection_generate_obs_timing():
    sensors = SensorCollection.fromYAML(yaml)

    ### SIMULATION TIMING
    t_start_mjd = Time("2011-11-11T11:11:11", format="isot", scale="utc").mjd
    t_end_mjd = t_start_mjd + 1

    assert sensors.obsTimes == None
    sensors.generate_obs_timing(t_start_mjd, t_end_mjd)
    assert sensors.obsTimes != None

    with pytest.raises(
        SensorException
    ):  # The observation schedule has already been generated.
        sensors.generate_obs_timing(t_start_mjd, t_end_mjd)


def test_add_sensor_post_timing():
    sensors = SensorCollection.fromYAML(yaml)

    ### SIMULATION TIMING
    t_start_mjd = Time("2011-11-11T11:11:11", format="isot", scale="utc").mjd
    t_end_mjd = t_start_mjd + 1

    SENSOR_PARAMS = {
        "lat": -21.89567,
        "lon": 114.0898731,
        "alt": 0.067845,
        "dra": 0.3,
        "ddec": 0.3,
        "obs_per_collect": 3,
        "obs_time_spacing": 1.5,
        "collect_gap_mean": 7200,
        "collect_gap_std": 800,
        "obs_limits": None,
        "id": "SENSOR",
    }

    assert sensors.numSensors == 11

    new_sensor_1 = madlib.GroundOpticalSensor(**SENSOR_PARAMS)
    sensors.add_sensor(new_sensor_1)
    assert sensors.numSensors == 12

    sensors.generate_obs_timing(t_start_mjd, t_end_mjd)

    new_sensor_2 = madlib.GroundOpticalSensor(**SENSOR_PARAMS)

    with pytest.raises(SensorException):
        sensors.add_sensor(new_sensor_2)

    assert sensors.numSensors == 12


def test_SensorCollection_observe(seed=seed):
    # --- generate an ObservationCollection
    # --- orbital state
    np.random.seed(seed)
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)

    # --- sensors
    sensors = SensorCollection.fromYAML(yaml)

    # --- setup sensor and generate ObservationCollection(s)
    sensors.generate_obs_timing(epoch, epoch + 1)  # + 1 [days]

    observations = sensors.observe(sat0)
    assert isinstance(observations, madlib._observation.ObservationCollection)
    assert isinstance(observations.pos_observed[0], madlib._observation.Observation)
    assert isinstance(observations.pos_truth[0], madlib._observation.Observation)


def test_SensorCollection_observe_student_dist(seed=seed):
    # --- generate an ObservationCollection with different noise distributions
    # --- orbital state
    np.random.seed(seed)
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)

    # --- sensors
    sensors = SensorCollection.fromYAML("tests/inputs/students_t_sensors.yaml")

    # --- setup sensor and generate ObservationCollection(s)
    sensors.generate_obs_timing(epoch, epoch + 1)  # + 1 [days]

    observations = sensors.observe(sat0)
    assert isinstance(observations, madlib._observation.ObservationCollection)
    assert isinstance(observations.pos_observed[0], madlib._observation.Observation)
    assert isinstance(observations.pos_truth[0], madlib._observation.Observation)


def test_SensorCollection_error(seed=seed):
    """Try and fail to make observations without generating obs timing"""
    # --- generate an ObservationCollection
    # --- orbital state
    np.random.seed(seed)
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)

    # --- sensors
    sensors = SensorCollection.fromYAML(yaml)

    failed = False
    try:
        observations = sensors.observe(sat0)
    except SensorException:
        failed = True

    assert failed


class TestInvalidYAML:
    """Group of tests for handling invalid YAML structures"""

    def test_missing_sensor_list(self):
        """Every sensor network YAML needs a top-level attribute called sensor_list."""
        failed = False
        try:
            params = SensorCollection.paramsFromYAML(
                "tests/inputs/invalid_sensor_1.yaml"
            )
        except MadlibException:
            failed = True

        assert failed

    def test_missing_required(self):
        """Make sure a YAML is invalid if one of its sensors is missing a required attribute."""
        failed = False
        try:
            params = SensorCollection.paramsFromYAML(
                "tests/inputs/invalid_sensor_2.yaml"
            )
        except MadlibException:
            failed = True

        assert failed

    def test_unknown_property(self):
        """Make sure a YAML is invalid if one of its sensors has an unexpected
        (in this case, misspelled) property."""
        failed = False
        try:
            params = SensorCollection.paramsFromYAML(
                "tests/inputs/invalid_sensor_3.yaml"
            )
        except MadlibException:
            failed = True

        assert failed



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Implementation of the Satellite class, which models the propagation of a
satellite (with or without a maneuver).
"""

import copy
from typing import Optional, Self

import astroforge as af
import numpy as np
from astropy import time
from numpy.typing import NDArray

from astroforge.coordinates import keplerian_to_cartesian
from astroforge.constants import GM as GM_Earth

from ._maneuver import ContinuousManeuver, ImpulsiveManeuver
from ._utils import MadlibException


class Satellite:
    """
    Satellite class for propagating a satellite (either with or without a maneuver).
    """

    _maneuver: Optional[ImpulsiveManeuver] | Optional[ContinuousManeuver] | None = (
        None  # contains maneuver information, if it exists
    )
    _epoch: float  # epoch of the satellite orbit (MJD, UTC)
    _xx: NDArray[np.float64]  # satellite position at epoch, shape = (3,)
    _vv: NDArray[np.float64]  # satellite velocity at epoch, shape = (3,)
    _aa: NDArray[np.float64]  # satellite anomalous acceleration at epoch, shape = (3,)

    # Optional TRUE orbit
    _epoch_true: float
    _xx_true: NDArray[np.float64]
    _vv_true: NDArray[np.float64]
    _aa_true: NDArray[np.float64]

    def __init__(
        self,
        epoch: float,
        pos: NDArray[np.float64],
        vel: NDArray[np.float64],
        acc: NDArray[np.float64] = np.zeros((3,)),
        maneuver_info: (
            Optional[ImpulsiveManeuver] | Optional[ContinuousManeuver] | None
        ) = None,
        epoch_true: float | None = None,
        pos_true: NDArray[np.float64] | None = None,
        vel_true: NDArray[np.float64] | None = None,
        acc_true: NDArray[np.float64] | None = None,
        **kwargs,
    ):
        """Instantiate a Satellite object.

        Parameters
        ----------
        epoch : float
            Timestamp (MJD, UTC) indicating the time at which this orbital state is defined
        pos : NDArray[np.float64]
            Vector describing the satellite position at epoch.
            Coordinate system is TETED. Units are km. Shape: (3,)
        vel : NDArray[np.float64]
            Vector describing the satellite velocity at epoch.
            Coordinate system is TETED. Units are km/s. Shape: (3,)
        acc : NDArray[np.float64], optional
            Vector describing the satellite anomalous acceleration at epoch.
            Coordinate system is TETED. Units are km/s/s. Shape: (3,), by default np.zeros((3,))
        maneuver_info : Optional[ImpulsiveManeuver], optional
            Maneuver object describing when and how the satellite performs an
            impulsive maneuver, by default None
        epoch_true : float | None, optional
            If not None, then this is the ACTUAL epoch of the orbit, and <epoch> is
            a reported value with some error. Units are MJD in UTC system. Default is None.
        pos_true : NDArray[np.float64] | None, optional
            If not None, then this vector is the satellite's ACTUAL position at epoch,
            and <pos> is a reported value with some error. Coordinate system is TETED. Units
            are km. Shape: (3,). Default is None.
        vel_true : NDArray[np.float64] | None, optional
            If not None, then this vector is the satellite's ACTUAL velocity at epoch,
            and <vel> is a reported value with some error. Coordinate system is TETED. Units
            are km/s. Shape: (3,). Default is None.
        acc_true : NDArray[np.float64] | None, optional
            If not None, then this vector is the satellite's ACTUAL anomalous acceleration at epoch,
            and <acc> is a reported value with some error. Coordinate system is TETED. Units
            are km/s/s. Shape: (3,). Default is None.

        Raises
        ------
        NotImplementedError
            kwargs are not yet supported
        """

        self.validate_input_vector(pos, "Position")
        self.validate_input_vector(vel, "Velocity")
        self.validate_input_vector(acc, "Anomalous Acceleration")

        self._epoch = epoch
        self._xx = pos
        self._vv = vel
        self._aa = acc
        self._maneuver = maneuver_info

        self.true_orbit_exists = False

        if epoch_true is not None:
            self.true_orbit_exists = True
            self.epoch_true = epoch_true
        else:
            self._epoch_true = epoch

        if pos_true is not None:
            self.true_orbit_exists = True
            self._xx_true = pos_true
        else:
            self._xx_true = pos

        if vel_true is not None:
            self.true_orbit_exists = True
            self._vv_true = vel_true
        else:
            self._vv_true = vel

        if acc_true is not None:
            self.true_orbit_exists = True
            self._aa_true = acc_true
        else:
            self._aa_true = acc

        if len(kwargs.items()) != 0:
            raise NotImplementedError("kwargs are not yet supported")

    @property
    def does_maneuver(self) -> bool:
        return self._maneuver is not None

    @property
    def maneuver(self) -> ImpulsiveManeuver | ContinuousManeuver | None:
        return self._maneuver

    @maneuver.setter
    def maneuver(self, val: ImpulsiveManeuver | ContinuousManeuver | None):
        self._maneuver = val

    @property
    def epoch(self) -> float:
        return self._epoch

    @epoch.setter
    def epoch(self, val: float):
        self._epoch = val

    @property
    def epoch_true(self) -> float:
        return self._epoch_true

    @epoch_true.setter
    def epoch_true(self, val: float):
        self.true_orbit_exists = True
        self._epoch_true = val

    @property
    def x(self) -> NDArray[np.float64]:
        return self._xx

    @x.setter
    def x(self, val: NDArray[np.float64]):
        self.validate_input_vector(val, "Position")
        self._xx = val

    @property
    def x_true(self) -> NDArray[np.float64] | None:
        return self._xx_true

    @x_true.setter
    def x_true(self, val: NDArray[np.float64]):
        self.validate_input_vector(val, "Position")
        self.true_orbit_exists = True
        self._xx_true = val

    @property
    def v(self) -> NDArray[np.float64]:
        return self._vv

    @v.setter
    def v(self, val: NDArray[np.float64]):
        self.validate_input_vector(val, "Velocity")
        self._vv = val

    @property
    def v_true(self) -> NDArray[np.float64]:
        return self._vv_true

    @v_true.setter
    def v_true(self, val: NDArray[np.float64]):
        self.validate_input_vector(val, "Velocity")
        self.true_orbit_exists = True
        self._vv_true = val

    @property
    def acc(self) -> NDArray[np.float64]:
        return self._aa

    @acc.setter
    def acc(self, val: NDArray[np.float64]):
        self.validate_input_vector(val, "Anomalous Acceleration")
        self._aa = val

    @property
    def acc_true(self) -> NDArray[np.float64]:
        return self._aa_true

    @acc_true.setter
    def acc_true(self, val: NDArray[np.float64]):
        self.validate_input_vector(val, "Anomalous Acceleration")
        self.true_orbit_exists = True
        self._aa_true = val

    def copy(self) -> Self:
        return copy.deepcopy(self)

    def validate_input_vector(self, input_vector, input_name):
        if not isinstance(input_vector, np.ndarray):
            raise TypeError(f"{input_name} must be a NumPy array")

        if input_vector.shape != (3,):
            raise ValueError(f"{input_name} must have shape (3,)")

    def propagate(
        self,
        times: float | NDArray[np.float64],
        ignore_maneuvers: bool = False,
        use_true_orbit: bool = False,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Propagates the satellite to the time(s) given.

        Parameters
        ----------
        times : float | NDArray[np.float64]
            Timestamp(s) for evaluating the satellite propagation (MJD, UTC)
        ignore_maneuvers : bool
            If True, propagate without applying any maneuvers, by default False.
        use_true_orbit : bool
            If True, and if input values were given for <epoch_true>, <pos_true>,
            <vel_true>, or <aa_true>, use these values for the true orbit to propagate
            instead of the reported values (<epoch>, <pos>, <vel>, <aa>)

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.float64]]
            First element of tuple is X : (N,3), position vector(s) at the time(s)
            requested. Coordinate system is TETED. Units are km.
            Second element of tuple is V : (N,3), velocity vector(s) at the time(s)
            requested, Coordinate system is TETED. Units are km/s.

        Raises
        ------
        ValueError
            One of the propagation times is the exact same time as the maneuver.
        """

        if isinstance(self.maneuver, ContinuousManeuver) and not ignore_maneuvers:
            raise MadlibException(
                "You cannot call Satellite.propagate() when Satellite._manuever is not type ImpulseManeuver. "
                "If you are trying to propagate a satellite with a ContinuousManeuver, use Satellite "
                "subclass ContinuousThrustSatellite."
            )

        if use_true_orbit and self.true_orbit_exists:
            epoch = self._epoch_true
            xx = self._xx_true
            vv = self._vv_true
            aa = self._aa_true
        else:
            epoch = self._epoch
            xx = self._xx
            vv = self._vv
            aa = self._aa

        scalar_input = False
        if isinstance(times, (float, int)):
            scalar_input = True
            times = np.asarray([times])

        # If the epoch is included in the propagation times, it could cause an
        # error in the numerical solver. Slightly perturb any such cases.
        times[times == epoch] += 1e-8

        if (
            (not self.does_maneuver)
            or (ignore_maneuvers)
            or (self.maneuver is not None and self.maneuver.time > times.max())
        ):
            # first propagate from the epoch to the first requested time
            T = np.array([epoch, times[0]])
            X, V = af.propagators.mp_srp(xx, vv, aa, T)
            X0, V0 = X[1], V[1]

            # no need to propagate further if only one time was provided
            if scalar_input or len(times) == 1:
                return (
                    X0[np.newaxis, :],
                    V0[np.newaxis, :],
                )  # Reshape needed for consistency

            # because there isn't a maneuver we can just propagate to all of the times all at once
            X, V = af.propagators.mp_srp(X0, V0, aa, times)
            return X, V
        else:
            # satellite does a maneuver sometime within the requested times

            # save the maneuver info, check for requested times at the same time as the maneuver
            mantime = self.maneuver.time  # type: ignore
            mandv = self.maneuver.dv  # type: ignore
            if mantime in times:
                raise ValueError(
                    "One of the propagation times is the exact same time as the maneuver."
                )

            # list of pos/vels from the various time segments
            Xkeep, Vkeep = [], []

            # propagate from the epoch through all of the requested pre-maneuver
            # times (if exist), and finally to the maneuver time
            T = np.hstack([epoch, times[times < mantime], mantime])
            X, V = af.propagators.mp_srp(xx, vv, aa, T)
            Xkeep.append(X[1:-1])
            Vkeep.append(V[1:-1])

            # apply delta-v at the time of the maneuver
            # first rotate the maneuver into ECI space
            x, v = X[-1], V[-1]
            R = self.rv2rsw(x, v)
            dv_eci = R.T @ mandv
            Xpm = x
            Vpm = v + dv_eci

            # propagate from the maneuver time to the rest of the requsted times
            # Note: there should be times after the maneuver time, otherwise we would have
            # entered the if above (self.maneuver.time > times.max())
            T = np.hstack([mantime, times[times > mantime]])
            X, V = af.propagators.mp_srp(Xpm, Vpm, aa, T)
            Xkeep.append(X[1:])
            Vkeep.append(V[1:])

            # stack the pos/vels to be a single array
            X = np.vstack(Xkeep)
            V = np.vstack(Vkeep)

            return X, V

    def create_cross_tag(
        self,
        cross_mjd: float,
        delta_pos_km: NDArray[np.float64],
        delta_vel_kms: NDArray[np.float64],
    ):
        """Create a satellite that will cross nearby this satellite,
        potentially creating a series of misattribution events.

        Args:
            cross_mjd : float
                A time (in MJD) at which the two objects will be nearby
            delta_pos_km : NDArray[np.float64]
                A vector representing the distance between the two satellites
                at time <cross_mjd> (ECI, km)
            delta_vel_kms : NDArray[np.float64]
                A vector representing the difference in the two satellites'
                velocities at time <cross_mjd> (ECI, km/s)

        Returns:
            Satellite
                The Satellite object for the crossing satellite, with epoch
                set at <cross_mjd>
        """

        # If the cross time is close to the epoch time, just use the epoch pos/vel
        if np.isclose(cross_mjd, self.epoch, rtol=0, atol=1e-6):
            x = self.x
            v = self.v
        else:
            # Propagate the satellite's initial orbit to the cross_mjd
            x, v = self.propagate(cross_mjd, ignore_maneuvers=True)
            # Satellite expects position as shape (3,) but x and v will be (1,3)
            x = x[0]
            v = v[0]

        # Add offsets.
        x_cross = x + delta_pos_km
        v_cross = v + delta_vel_kms

        cross_sat = Satellite(epoch=cross_mjd, pos=x_cross, vel=v_cross)

        return cross_sat

    @staticmethod
    def rv2rsw(r: NDArray[np.float64], v: NDArray[np.float64]) -> NDArray[np.float64]:
        """Compute rotation matrix from ECI frame to RSW frame.

        Parameters
        ----------
        r : NDArray[np.float64]
            Position vector (ECI, km)
        v : NDArray[np.float64]
            Velocity vector (ECI, km/s)

        Returns
        -------
        NDArray[np.float64]
            Rotation matrix from ECI frame to RSW frame
        """

        unit = lambda z: z / np.sqrt(z @ z)

        # get unit vector in each dimension
        # radial
        rvec = unit(r)

        # cross-track
        wvec = np.cross(r, v)
        wvec = unit(wvec)

        # along-track
        svec = np.cross(wvec, rvec)
        svec = unit(svec)

        # create rotation matrix
        T = np.zeros((3, 3))
        T[0, :] = rvec
        T[1, :] = svec
        T[2, :] = wvec

        return T

    @classmethod
    def from_GEO_longitude(cls, lon: float, epoch: float) -> Self:
        """Create a Satellite in GEO at the given longitude

        Parameters
        ----------
        lon : float
            Longitude of the GEO satellite (deg)
        epoch : float
            Timestamp (MJD, UTC) indicating the time at which this orbital state is defined

        Returns
        -------
        Self
            Instance of the Satellite class
        """

        lon = np.deg2rad(lon)

        # note: this only works for GEO because GEO has zero velocity in ITRS
        x_itrs = af.constants.Rgeo * np.array([np.cos(lon), np.sin(lon), 0.0])
        x, v = af.coordinates.PosVelConversion(
            af.coordinates.ITRSToTETED, epoch, x_itrs, np.zeros(3)
        )
        return cls(epoch, x, v)

    @classmethod
    def from_keplerian(
        cls,
        epoch: float,
        inclination_rad: float,
        raan_rad: float,
        argp_rad: float,
        ecc: float,
        semi_major_axis_km: float,
        mean_anomaly_rad: float,
        GM: float = GM_Earth,
    ) -> Self:
        """Create a Satellite from Keplerian elements

        Parameters
        ----------
        epoch : float
            Timestamp (MJD, UTC) indicating the time at which this orbital state is defined
        inclination_rad : float
            Orbital inclination in radians
        raan_rad : float
            Right ascension of the ascending node in radians
        argp_rad : float
            Argument of pericenter in radians
        ecc : float
            Eccentricity, a number between 0 and 1
        semi_major_axis_km : float
            Semi-major axis in km
        mean_anomaly_rad : float
            Mean anomaly in radians
        GM : float, optional
            Standard gravitational parameter (the product of the gravitational constant and the mass of a given astronomical body such as the Sun or Earth) in km^3/s^2. Earth's standard gravitational parameter by default (398600.4418 km^3/s^2)

        Returns
        -------
        Self
            Instance of the Satellite class
        """

        x, v = keplerian_to_cartesian(
            inclination_rad,
            raan_rad,
            argp_rad,
            ecc,
            semi_major_axis_km,
            mean_anomaly_rad,
            GM,
        )

        # Reshape from (1,3) to (3,)
        x = x.ravel()
        v = v.ravel()

        return cls(epoch, x, v)

    def __str__(self):
        epochstr = time.Time(self.epoch, format="mjd").isot
        s = ""
        s += "Satellite:\n"
        s += f"  epoch = {epochstr}\n"
        s += f"  x = {self.x.tolist()}\n"
        s += f"  v = {self.v.tolist()}\n"
        s += f"  acc = {self.acc.tolist()}"
        return s

    def __repr__(self):
        return self.__str__()


class ContinuousThrustSatellite(Satellite):
    """Satellite subclass for propagating an orbit that uses continuous thrust (i.e. ContinuousManeuver)"""

    def __init__(
        self,
        epoch: float,
        pos: NDArray[np.float64],
        vel: NDArray[np.float64],
        acc: NDArray[np.float64] = np.zeros((3,)),
        maneuver_info: Optional[ContinuousManeuver] | None = None,
        epoch_true: float | None = None,
        pos_true: NDArray[np.float64] | None = None,
        vel_true: NDArray[np.float64] | None = None,
        acc_true: NDArray[np.float64] | None = None,
        **kwargs,
    ):
        super().__init__(
            epoch,
            pos,
            vel,
            acc,
            maneuver_info,
            epoch_true=epoch_true,
            pos_true=pos_true,
            vel_true=vel_true,
            acc_true=acc_true,
        )

        self._propfun = self._setup_propfun()

    def propagate(
        self,
        times: float | NDArray[np.float64],
        ignore_maneuvers: bool = False,
        use_true_orbit: bool = False,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Top level continuous thrust propagate method. Propagates the satellite to the time(s) given.

        Parameters
        ----------
        times : float | NDArray[np.float64]
            Timestamp(s) for evaluating the satellite propagation (MJD, UTC)
        ignore_maneuvers : bool
            If True, propagate without applying any maneuvers, by default False.
        use_true_orbit : bool
            If True, and if input values were given for <epoch_true>, <pos_true>,
            <vel_true>, or <aa_true>, use these values for the true orbit to propagate
            instead of the reported values (<epoch>, <pos>, <vel>, <aa>)

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.float64]]
            First element of tuple is X : (N,3), position vector(s) at the time(s)
            requested. Coordinate system is TETED. Units are km.
            Second element of tuple is V : (N,3), velocity vector(s) at the time(s)
            requested, Coordinate system is TETED. Units are km/s.
        """

        scalar_input = False
        if isinstance(times, (float, int)):
            scalar_input = True
            times = np.asarray([times])

        if ignore_maneuvers:
            return super().propagate(
                times,
                ignore_maneuvers=True,
                use_true_orbit=use_true_orbit,
            )

        if use_true_orbit and self.true_orbit_exists:
            epoch = self._epoch_true
            xx = self._xx_true
            vv = self._vv_true
            aa = self._aa_true
        else:
            epoch = self._epoch
            xx = self._xx
            vv = self._vv
            aa = self._aa

        # do an initial propagation to the first requested time, that way
        # we can ensure the propagation need only go one direction
        T = np.array([epoch, times[0]])
        X, V = self._propfun(xx, vv, aa, T)
        X0, V0 = X[1], V[1]

        # no need to propagate further if only one time was provided
        if scalar_input or len(times) == 1:
            return X0, V0

        X, V = self._propfun(X0, V0, aa, times)
        return X, V

    def _setup_propfun(self):
        # the static force model used in numerical integration; this does not include the continuous thrust
        self._fm = af.force_models.F_mp_srp

        def full_eom(t: float, xxdot: NDArray[np.float64]) -> NDArray[np.float64]:
            """Full equations of motion (EOM) to be used by the propagator's ODE solver

            Parameters
            ----------
            t : float
                Timestamp for evaluating the satellite propagation
            xxdot : NDArray[np.float64]
                Position, velocity, & acceleration state vectors concatenated
                into a single 1D array, shape: (9,)

            Returns
            -------
            NDArray[np.float64]
                The derivative, velocity, acceleration and jerk concatenated into a single
                array of shape: (9,). Acceleration due to continuous thrust acceleration
                added, if applicable.
            """
            dx = self._fm(t, xxdot)
            if self.maneuver is not None and callable(self.maneuver):
                dx[3:6] += self.maneuver(t / 86400.0)

            return dx

        def propfun(
            x0: NDArray[np.float64],
            v0: NDArray[np.float64],
            alpha: NDArray[np.float64],
            T: NDArray[np.float64],
        ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
            """ContinuousThrustSatellite object propagation function

            Parameters
            ----------
            x0 : NDArray[np.float64]
                Position vector, shape: (3,)
            v0 : NDArray[np.float64]
                Velocity vector, shape: (3,)
            alpha : NDArray[np.float64]
                Acceleration vector, shape: (3,)
            T : NDArray[np.float64]
                Array of timestamps (MJD, UTC) to propagate to, shape: (N,)

            Returns
            -------
            tuple[NDArray[np.float64], NDArray[np.float64]]
                First element of tuple is X : (N,3), position vector(s) at the time(s)
                requested.
                Second element of tuple is V : (N,3), velocity vector(s) at the time(s)
                requested.
            """
            y0 = np.hstack([x0, v0, alpha])
            T = T.copy()
            T += af.coordinates.dut1utc(T) / 86400.0
            T *= 86400.0

            out = af.propagators.propagator(full_eom, y0, T, atol=1e-9, rtol=1e-9)

            X = out[:, :3]
            V = out[:, 3:6]
            return (X, V)

        return propfun



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import Protocol, Tuple

import astroforge.coordinates as afc
import numpy as np
from astroforge import R_sun, R_earth
from numpy.typing import NDArray

from madlib._utils import MadlibException
from ._observation import Observation, ObservationCollection
from ._satellite import Satellite
from ._utils import calc_separation_angle


class _Sensor(Protocol):
    """Class is not yet implemented"""

    @abstractmethod
    def generate_obs_timing(self, start: float, end: float) -> NDArray[np.float64]:
        """Given a start time and an end time (in MJD) as well as the sensor's
        defined parameters, generate an array of observation times (also in MJD).

        Parameters
        ----------
        start : float
            Earliest possible observation timestamp (in MJD)
        end : float
            Latest possible observation timestamp (in MJD)

        Returns
        -------
        NDArray[np.float64]
            An array of times (in MJD) at which an observation will be made.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError

    @abstractmethod
    def observe(self, sat: Satellite, times: float | NDArray[np.float64]):
        """Observe a given satellite with this sensor at the times given.

        Parameters
        ----------
        sat : Satellite
            Satellite object to observe
        times : float | NDArray[np.float64]
            Time(s) to observe the satellite. Specified as MJD in the UTC system.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError


class _OpticalSensor(_Sensor):
    """Generic class for optical sensors"""

    id: str | None

    # site metric accuracy
    dra: float  # arcsec
    ddec: float  # arcsec

    # timing
    obs_per_collect: int | tuple[int, int]
    obs_time_spacing: float
    collect_gap_mean: float
    collect_gap_std: float

    # observation limits
    obs_limits: dict | None

    def __init__(
        self,
        dra: float,
        ddec: float,
        collect_gap_mean: float,
        obs_limits: dict | None = None,
        collect_gap_std: float = 0.0,
        obs_per_collect: int | tuple[int, int] = 1,
        obs_time_spacing: float = 1.0,
        students_dof: int | None = None,
        id: str | None = None,
        cross_tag: Satellite | None = None,
        cross_tag_limit_arcsec: float = 100.0,
    ):
        """Construct an _OpticalSensor object

        Parameters
        ----------
        dra : float
            Metric accuracy in the right ascension direction
        ddec : float
            Metric accuracy in the declination direction
        collect_gap_mean : float
            Average time (seconds) between collects on a satellite
        obs_limits : dict | None, optional
            Dictionary of the limits on what this sensor can observe. This can be
            obvious (e.g. can't observe through the Earth), or based on sensor
            sensitivity (e.g. range limit for radar), by default None
        collect_gap_std : float, optional
            Standard deviation of times (seconds) between collects on a satellite, by default 0.0
        obs_per_collect : int | tuple[int, int], optional
            Typical number of observations per collect. This can be a constant, or
            a tuple of (min, max) for randomly sampling, by default 1
        obs_time_spacing : float, optional
            Time between observations within a collect, by default 1.0
        id : str | None, optional
            Unique Sensor ID, by default None
        cross_tag : madlib.Satellite | None, optional
            An optional nearby satellite that the sensor has misattributed
            as the target, by default None (no misattribution)
        cross_tag_limit_arcsec : float, optional
            The maximum allowed separation (in arcseconds) between the target's
            expected position and the cross_tag satellite. At further separations,
            observations of the misattributed satellite will be ignored.
            By default 100.0
        students_dof : int | None, optional
            Degrees of freedom to use for T-distribution when generating noise, by default None (Gaussian)
        """

        self.dra = dra
        self.ddec = ddec
        self.obs_per_collect = obs_per_collect
        self.obs_time_spacing = obs_time_spacing
        self.collect_gap_mean = collect_gap_mean
        self.collect_gap_std = collect_gap_std
        self.obs_limits = obs_limits
        self.id = id
        self.cross_tag = cross_tag
        self.cross_tag_limit_arcsec = cross_tag_limit_arcsec
        self.students_dof = students_dof

    def generate_obs_timing(self, start: float, end: float) -> NDArray[np.float64]:
        """Randomly generate a realistic time sampling of observations.

        Parameters
        ----------
        start : float
            Timestamp (MJD) corresponding to the beginning of the observing period
        end : float
            Timestamp (MJD) corresponding to the end of the observing period

        Returns
        -------
        NDArray[np.float64]
            Array of MJD timestamps corresponding to typical observing times on a satellite

        Raises
        ------
        ValueError
            Start timestamp must be before end timestamp
        """

        if start >= end:
            raise ValueError(f"start ({start}) must be before end ({end})")

        def gap_gen() -> float:
            return self.collect_gap_mean + self.collect_gap_std * np.random.randn()

        istuple = isinstance(self.obs_per_collect, (tuple, list))

        def nobs_gen() -> int:
            if istuple:
                return np.random.randint(
                    self.obs_per_collect[0], self.obs_per_collect[1] + 1  # type: ignore
                )
            else:
                return self.obs_per_collect  # type: ignore

        # --- generate collect times
        collect_times = []
        latest = start
        first = True
        while latest < end:
            # generate a gap time at random
            gap = gap_gen() / 86400

            # if this is the first gap, randomly scale it between 0-1
            # (otherwise the first observation will be biased late)
            if first:
                gap *= np.random.rand()
                first = False

            # add the gap time to the latest collect time, and append to the
            # list if it isn't too late
            latest += gap
            if latest < end:
                collect_times.append(latest)

        # --- generate observation times within each collect
        obs_times = []
        for coll in collect_times:
            # generate a (possibly) random number of observations in this collect
            nobs = nobs_gen()

            # compute the time of each observation given the collect start time
            times = coll + (np.arange(nobs) * self.obs_time_spacing) / 86400

            # handle edge effect at the end of the observing period
            # all observations in this collect must be before the end, otherwise
            # none are saved
            if (times < end).all():
                obs_times.append(times)

        if len(obs_times) > 0:
            obs_times = np.hstack(obs_times)
        else:
            obs_times = np.array([])
        return obs_times

    def validate_limits(self, obs: Observation) -> bool:
        """Determine whether Observation is possible based on the sensor limits.

        Parameters
        ----------
        obs : Observation
            madlib.Observation class for holding observables

        Returns
        -------
        bool
            Whether or not the Observation is possible
        """

        if not self.obs_limits:
            return True

        flag = True
        obsdict = obs.__dict__
        for key, val in self.obs_limits.items():
            if key in obsdict:
                if obsdict[key] is not None:
                    flag = flag and obsdict[key] > val[0] and obsdict[key] < val[1]
                else:
                    msg = f"ERROR: Invalid observation, value of {key} is None"
                    raise MadlibException(msg)
            else:
                msg = (
                    f"ERROR: {key} is an invalid variable for observation limits. "
                    "Check the YAML configuration for sensor {self.id}"
                )
                raise MadlibException(msg)

        return flag


class GroundOpticalSensor(_OpticalSensor):
    """Class for modeling the observing of a satellite with an optical sensor"""

    # site geography
    _site_reported_itrs: NDArray[np.float64]
    lat: float
    lon: float
    alt: float

    # site metric accuracy
    dra: float  # arcsec
    ddec: float  # arcsec

    # timing
    obs_per_collect: int | tuple[int, int]
    collect_gap_mean: float
    collect_gap_std: float

    # observation limits
    obs_limits: dict | None

    def __init__(
        self,
        lat: float,
        lon: float,
        alt: float,
        dra: float,
        ddec: float,
        collect_gap_mean: float,
        obs_limits: dict | None = None,
        collect_gap_std: float = 0.0,
        obs_per_collect: int | tuple[int, int] = 1,
        obs_time_spacing: float = 1.0,
        id: str | None = None,
        lat_truth: float | None = None,
        lon_truth: float | None = None,
        alt_truth: float | None = None,
        cross_tag: Satellite | None = None,
        cross_tag_limit_arcsec: float = 100.0,
        **extras,
    ):
        """Construct a GroundOpticalSensor object

        Parameters
        ----------
        lat : float
            Site geodetic latitude (deg)
        lon : float
            Site geodetic longitude (deg)
        alt : float
            Site altitude above WGS-84 reference ellipsoid
        dra : float
            Metric accuracy in the right ascension direction
        ddec : float
            Metric accuracy in the declination direction
        collect_gap_mean : float
            Average time (seconds) between collects on a satellite
        obs_limits : dict | None, optional
            Dictionary of the limits on what this sensor can observe. This can be
            obvious (e.g. can't observe through the Earth), or based on sensor
            sensitivity (e.g. range limit for radar), by default None
        collect_gap_std : float, optional
            Standard deviation of times (seconds) between collects on a satellite, by default 0.0
        obs_per_collect : int | tuple[int, int], optional
            Typical number of observations per collect. This can be a constant, or
            a tuple of (min, max) for randomly sampling, by default 1
        obs_time_spacing : float, optional
            Time between observations within a collect, by default 1.0
        id : str | None, optional
            Unique Sensor ID, by default None
        lat_truth : float | None, optional
            If not None, the latitude specified by <lat> is reported by
            the sensor as its position, but it is incorrect. The actual
            position is specified by <lat_truth>. By default None.
        lon_truth : float | None, optional
            If not None, the longitude specified by <lon> is reported by
            the sensor as its position, but it is incorrect. The actual
            position is specified by <lon_truth>. By default None.
        alt_truth : float | None, optional
            If not None, the altitude specified by <alt> is reported by
            the sensor as its position, but it is incorrect. The actual
            position is specified by <alt_truth>. By default None.
        cross_tag : madlib.Satellite | None, optional
            An optional nearby satellite that the sensor has misattributed
            as the target, by default None (no misattribution)
        cross_tag_limit_arcsec : float, optional
            The maximum allowed separation (in arcseconds) between the target's
            expected position and the cross_tag satellite. At further separations,
            observations of the misattributed satellite will be ignored.
            By default 100.0
        """

        super().__init__(
            dra=dra,
            ddec=ddec,
            collect_gap_mean=collect_gap_mean,
            obs_limits=obs_limits,
            collect_gap_std=collect_gap_std,
            obs_per_collect=obs_per_collect,
            obs_time_spacing=obs_time_spacing,
            id=id,
            cross_tag=cross_tag,
            cross_tag_limit_arcsec=cross_tag_limit_arcsec,
        )

        self.lat = lat
        self.lon = lon
        self.alt = alt
        self._site_reported_itrs = afc.LatLonAltToITRS(lat, lon, alt)

        self.lat_truth = lat
        self.lon_truth = lon
        self.alt_truth = alt

        if lat_truth is not None:
            self.lat_truth = lat_truth
        if lon_truth is not None:
            self.lon_truth = lon_truth
        if alt_truth is not None:
            self.alt_truth = alt_truth

        self._site_truth_itrs = afc.LatLonAltToITRS(
            self.lat_truth, self.lon_truth, self.alt_truth
        )

    def observe(
        self,
        target_satellite: Satellite,
        times: float | NDArray[np.float64] | Tuple[float, float],
    ) -> ObservationCollection:
        """Observe a satellite with this sensor model at the times given. Observations
        are computed in three forms:
            - truth: Measurements that would be returned if there were no random or
                systematic errors and knowledge of the orbit and maneuvers was perfect.
                These are the observations you would see in a perfect world.
            - expected: Measurements without random noise, but also without maneuver knowledge
                and with any systematic errors on sensor position and/or satellite orbit.
                These are the observations that you THINK you should get, given your actual knowledge
                of the system.
            - measured: The actual output of the simulated system, including all random and
                systematic sources of error and following the target satellite's full trajectory.
                This can also contain cross-tag events.

        Parameters
        ----------
        target_satellite : Satellite
            madlib.Satellite object to observe
        times : float | NDArray[np.float64] | Tuple[float, float]
            Time(s) to observe the satellite. Specified as MJD in the UTC system.
                - If a single float is given, observation will be made at just that time.
                - If a numpy array is given, observations will be made at each time in the array.
                - If a tuple of two floats is given, they will represent the start and end
                  of observations, and observation times will be generated accordingly

        Returns
        -------
        ObservationCollection
            A container object holding a realistic observation of the satellite
            given the sensor noise parameters and the "true" observation that
            excludes all noise sources.
        """

        # Handle scalar and tuple time inputs
        if isinstance(times, (float, int)):
            times = np.asarray([times])

        if isinstance(times, tuple):
            start_mjd = min(times)
            end_mjd = max(times)
            times = self.generate_obs_timing(start_mjd, end_mjd)

        num_obs = len(times)

        # Return empty ObservationCollection if there are no observation times
        if num_obs == 0:
            pos_observed = np.array([])
            pos_truth = np.array([])
            pos_expected = np.array([])

            return ObservationCollection(
                pos_observed=pos_observed,
                pos_truth=pos_truth,
                pos_expected=pos_expected,
            )

        # Propagate satellite to the desired times, with and without maneuvers
        x_target, v_target = target_satellite.propagate(times, use_true_orbit=True)
        x_target_expected, v_target_expected = target_satellite.propagate(
            times,
            ignore_maneuvers=True,
        )

        # Coordinate conversion to observables
        # (observables for an optical sensor are RA/Dec)
        # First rotate the site's reported and true locations into TETED
        x_site_reported, v_site_reported = self._site_loc_TETED(times)
        x_site_truth, v_site_truth = self._site_loc_TETED(times, use_true_location=True)

        # Now compute the true and expected RA/Dec, range, and range_rate
        (fp_state_truth, r_truth, r_rate_truth) = afc.PosVelToFPState(
            x_target, v_target, x_site_truth, v_site_truth
        )
        ra_truth = fp_state_truth[:, 0] * 180 / np.pi
        dec_truth = fp_state_truth[:, 1] * 180 / np.pi
        az_truth, el_truth = self._eci_to_az_el(x_target, times)

        (fp_state_expected, r_expected, r_rate_expected) = afc.PosVelToFPState(
            x_target_expected,
            v_target_expected,
            x_site_reported,
            v_site_reported,
        )
        ra_expected = fp_state_expected[:, 0] * 180 / np.pi
        dec_expected = fp_state_expected[:, 1] * 180 / np.pi
        az_expected, el_expected = self._eci_to_az_el(x_target_expected, times)

        # Compute solar elevation at observation times (assumed invariant to expectation (small errors))
        x_sun = np.array([R_sun(_t) for _t in times])
        az_sun, el_sun = self._eci_to_az_el(x_sun, times)

        # Collect into true (what you SHOULD see) and expected (what you THINK you'll see) observations
        obs_truth: np.ndarray[Observation, np.dtype[np.float64]] = np.array(
            [
                Observation(
                    mjd=times[n],
                    ra=ra_truth[n],
                    dec=dec_truth[n],
                    az=az_truth[n],
                    el=el_truth[n],
                    sun_el=el_sun[n],
                    sensor_id=self.id,
                )
                for n in range(num_obs)
            ]
        )

        obs_expected: np.ndarray[Observation, np.dtype[np.float64]] = np.array(
            [
                Observation(
                    mjd=times[n],
                    ra=ra_expected[n],
                    dec=dec_expected[n],
                    az=az_expected[n],
                    el=el_expected[n],
                    sun_el=el_sun[n],
                    sensor_id=self.id,
                )
                for n in range(num_obs)
            ]
        )

        ### Actual/Observed measurements

        # Apply misattributed observations (cross-tags), if requested
        if self.cross_tag is not None:
            x_target, v_target = self.cross_tag.propagate(times, use_true_orbit=True)

            # Coordinate conversion to observables using target's and sensor's true positions
            (fp_state_measured, r_measured, r_rate_measured) = afc.PosVelToFPState(
                x_target, v_target, x_site_truth, v_site_truth
            )
            ra_measured = fp_state_measured[:, 0] * 180 / np.pi
            dec_measured = fp_state_measured[:, 1] * 180 / np.pi

        # If this isn't a cross-tag, then the measurement is just the noise-addled truth
        else:
            r_measured = r_truth
            ra_measured = ra_truth
            dec_measured = dec_truth

        # Apply noise sources
        ra_delta = self.dra / 3600 * np.random.randn(num_obs)
        dec_delta = self.ddec / 3600 * np.random.randn(num_obs)
        ra_noisy = ra_measured + ra_delta
        dec_noisy = dec_measured + dec_delta

        x_noisy = np.zeros(x_target.shape)
        for n in range(num_obs):
            x_noisy[n] = x_site_truth[n] + r_measured[n] * spherical_to_cartesian(
                ra_noisy[n], dec_noisy[n]
            )

        az_noisy, el_noisy = self._eci_to_az_el(x_noisy, times)

        # Collect actual observables (what you DO see)
        obs_measured: np.ndarray[Observation, np.dtype[np.float64]] = np.array(
            [
                Observation(
                    mjd=times[n],
                    ra=ra_noisy[n],
                    dec=dec_noisy[n],
                    az=az_noisy[n],
                    el=el_noisy[n],
                    range_=None,  # N/A for optical sensors
                    range_rate=None,  # N/A for optical sensors
                    sun_el=el_sun[n],
                    sensor_id=self.id,
                )
                for n in range(num_obs)
            ]
        )

        # Remove observations that are not within the sensor limits
        valid = np.array([self.validate_limits(x) for x in obs_truth])
        obs_measured = obs_measured[valid]
        obs_truth = obs_truth[valid]
        obs_expected = obs_expected[valid]

        observations = ObservationCollection(
            pos_observed=obs_measured,
            pos_truth=obs_truth,
            pos_expected=obs_expected,
        )

        return observations

    def _eci_to_az_el(
        self, x: NDArray[np.float64], mjd: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Convert TETED positions to Az/El from the site.

        Parameters
        ----------
        x : NDArray[np.float64]
            TETED positions
        mjd : NDArray[np.float64]
            timestamps (mjd)

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.float64]]
            Az/El results
        """

        # rotate TETED -> ITRS -> SEZ -> Az/El
        out = []
        for n in range(len(x)):
            # breakpoint()
            xitrs = afc.TETEDToITRS(mjd[n], x[n])
            xsez = afc.ITRSToSEZ(xitrs, self._site_reported_itrs, self.lat, self.lon)
            az, el, _ = afc.SEZToAzElRange(xsez)
            out.append([az, el])

        out = np.vstack(out)
        return out[:, 0], out[:, 1]

    def _site_loc_TETED(
        self,
        times: NDArray[np.float64],
        use_true_location: bool = False,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Rotate the site location to TETED for a given set of times.

        Parameters
        ----------
        times : NDArray[np.float64]
            timestamps at which to compute the site location
        use_true_location : bool, optional
            If True, use the sensor's true position (as opposed to reported)
            in the calculation. By default False.

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.float64]]
            site locations for timestamps
        """

        if use_true_location:
            site_itrs = self._site_truth_itrs
        else:
            site_itrs = self._site_reported_itrs

        temp = np.vstack(
            [
                afc.PosVelConversion(
                    afc.ITRSToTETED, times[n], site_itrs, np.zeros((3,))
                )
                for n in range(len(times))
            ]
        )
        xsite = temp[0::2]
        vsite = temp[1::2]

        return xsite, vsite


class SpaceOpticalSensor(_OpticalSensor):
    """Class for modeling observations with an optical sensor in Earth orbit."""

    # site metric accuracy
    dra: float  # arcsec
    ddec: float  # arcsec

    # timing
    obs_per_collect: int | tuple[int, int]
    collect_gap_mean: float
    collect_gap_std: float

    # observation limits
    obs_limits: dict | None

    def __init__(
        self,
        sensor_satellite: Satellite,
        dra: float,
        ddec: float,
        collect_gap_mean: float,
        sensor_satellite_truth: Satellite | None = None,
        obs_limits: dict | None = None,
        collect_gap_std: float = 0.0,
        obs_per_collect: int | tuple[int, int] = 1,
        obs_time_spacing: float = 1.0,
        id: str | None = None,
        cross_tag: Satellite | None = None,
        cross_tag_limit_arcsec: float = 100.0,
        **extras,
    ):
        """Construct a SpaceOpticalSensor object

        Parameters
        ----------
        sensor_satellite : madlib.Satellite
            Satellite object governing sensor motion
        dra : float
            Metric accuracy in the right ascension direction
        ddec : float
            Metric accuracy in the declination direction
        collect_gap_mean : float
            Average time (seconds) between collects on a satellite
        sensor_satellite_truth: madlib.Satellite | None, optional
            If not None, this defines the *actual* orbit of the satellite, turning
            the <satellite> argument into the *expected* orbit. This will affect
            residual calculations. By default, None.
        obs_limits : dict | None, optional
            Dictionary of the limits on what this sensor can observe. This can be
            obvious (e.g. can't observe through the Earth), or based on sensor
            sensitivity (e.g. range limit for radar), by default None
        collect_gap_std : float, optional
            Standard deviation of times (seconds) between collects on a satellite, by default 0.0
        obs_per_collect : int | tuple[int, int], optional
            Typical number of observations per collect. This can be a constant, or
            a tuple of (min, max) for randomly sampling, by default 1
        obs_time_spacing : float, optional
            Time between observations within a collect, by default 1.0
        id : str | None, optional
            Unique Sensor ID, by default None
        cross_tag : madlib.Satellite | None, optional
            An optional nearby satellite that the sensor has misattributed
            as the target, by default None (no misattribution)
        cross_tag_limit_arcsec : float, optional
            The maximum allowed separation (in arcseconds) between the target's
            expected position and the cross_tag satellite. At further separations,
            observations of the misattributed satellite will be ignored.
            By default 100.0
        """

        super().__init__(
            dra=dra,
            ddec=ddec,
            collect_gap_mean=collect_gap_mean,
            obs_limits=obs_limits,
            collect_gap_std=collect_gap_std,
            obs_per_collect=obs_per_collect,
            obs_time_spacing=obs_time_spacing,
            id=id,
            cross_tag=cross_tag,
            cross_tag_limit_arcsec=cross_tag_limit_arcsec,
        )

        self.sensor_satellite = sensor_satellite
        self.sensor_satellite_truth = sensor_satellite_truth

    def observe(
        self,
        target_satellite: Satellite,
        times: float | NDArray[np.float64] | Tuple[float, float],
    ):
        """Observe a satellite with this sensor model at the times given. Observations
        are computed in three forms:
            - truth: Measurements that would be returned if there were no random or
                systematic errors and knowledge of the target's orbit and maneuvers was perfect.
                These are the observations you would see in a perfect world.
            - expected: Measurements without random noise, but also without maneuver knowledge
                and with any systematic errors on sensor position and/or satellite orbit.
                These are the observations that you THINK you should get, given your actual knowledge
                of the system.
            - measured: The actual output of the simulated system, including all random and
                systematic sources of error and following the target satellite's full trajectory.
                This can also contain cross-tag events.

        Parameters
        ----------
        target_satellite : Satellite
            madlib.Satellite object to observe
        times : float | NDArray[np.float64] | Tuple[float, float]
            Time(s) to observe the satellite. Specified as MJD in the UTC system.
                - If a single float is given, observation will be made at just that time.
                - If a numpy array is given, observations will be made at each time in the array.
                - If a tuple of two floats is given, they will represent the start and end
                  of observations, and observation times will be generated accordingly

        Returns
        -------
        ObservationCollection
            A container object holding a realistic observation of the satellite
            given the sensor noise parameters and the "true" observation that
            excludes all noise sources.
        """

        # Handle scalar and tuple time inputs
        if isinstance(times, (float, int)):
            times = np.asarray([times])

        if isinstance(times, tuple):
            start_mjd = min(times)
            end_mjd = max(times)
            times = self.generate_obs_timing(start_mjd, end_mjd)

        num_obs = len(times)

        # Return empty ObservationCollection if there are no observation times
        if num_obs == 0:
            pos_observed = np.array([])
            pos_truth = np.array([])
            pos_expected = np.array([])

            return ObservationCollection(
                pos_observed=pos_observed,
                pos_truth=pos_truth,
                pos_expected=pos_expected,
            )

        # Propagate sensor expected position to the desired times
        x_sensor_reported, v_sensor_reported = self.sensor_satellite.propagate(times)

        # If a true orbit has been given, propagate it as well
        if self.sensor_satellite_truth is not None:
            x_sensor_truth, v_sensor_truth = self.sensor_satellite_truth.propagate(
                times
            )
        else:
            x_sensor_truth = x_sensor_reported
            v_sensor_truth = v_sensor_reported

        # Propagate satellite's true and expected orbits to the desired times
        x_target, v_target = target_satellite.propagate(times, use_true_orbit=True)
        x_target_expected, v_target_expected = target_satellite.propagate(
            times, ignore_maneuvers=True
        )

        # Compute true and expected RA/Dec, range, and range_rate
        (fp_state_truth, r_truth, r_rate_truth) = afc.PosVelToFPState(
            x_target,
            v_target,
            x_sensor_truth,
            v_sensor_truth,
        )
        ra_truth = fp_state_truth[:, 0] * 180 / np.pi
        dec_truth = fp_state_truth[:, 1] * 180 / np.pi

        (fp_state_expected, r_expected, r_rate_expected) = afc.PosVelToFPState(
            x_target_expected,
            v_target_expected,
            x_sensor_reported,
            v_sensor_reported,
        )
        ra_expected = fp_state_expected[:, 0] * 180 / np.pi
        dec_expected = fp_state_expected[:, 1] * 180 / np.pi

        # Compute solar separation angle at observation times
        # (using the sensor's true orbit)
        x_sun = np.array([R_sun(_t) for _t in times])
        sun_angle_deg = calc_separation_angle(
            x_sun - x_sensor_truth,
            x_target - x_sensor_truth,
            in_deg=True,
        )

        # Collect into true (what you SHOULD see) and expected (what you THINK you'll see) observations
        obs_truth: np.ndarray[Observation, np.dtype[np.float64]] = np.array(
            [
                Observation(
                    mjd=times[n],
                    ra=ra_truth[n],
                    dec=dec_truth[n],
                    sun_separation=sun_angle_deg[n],
                    sensor_id=self.id,
                )
                for n in range(num_obs)
            ]
        )

        obs_expected: np.ndarray[Observation, np.dtype[np.float64]] = np.array(
            [
                Observation(
                    mjd=times[n],
                    ra=ra_expected[n],
                    dec=dec_expected[n],
                    sun_separation=sun_angle_deg[n],
                    sensor_id=self.id,
                )
                for n in range(num_obs)
            ]
        )

        ### Actual/Observed Measurements

        # Apply misattributed observations, if requested
        if self.cross_tag is not None:
            x_target, v_target = self.cross_tag.propagate(times, use_true_orbit=True)

            # Compute (observed) RA/Dec, range, and range_rate
            (fp_state_obs, r_obs, r_rate_obs) = afc.PosVelToFPState(
                x_target, v_target, x_sensor_truth, v_sensor_truth
            )
            ra_measured = fp_state_obs[:, 0] * 180 / np.pi
            dec_measured = fp_state_obs[:, 1] * 180 / np.pi

        # If this isn't a cross-tag, then the measurement is just the noise-addled truth
        else:
            ra_measured = ra_truth
            dec_measured = dec_truth

        # Apply noise sources
        ra_delta = self.dra / 3600 * np.random.randn(num_obs)
        dec_delta = self.ddec / 3600 * np.random.randn(num_obs)
        ra_noisy = ra_measured + ra_delta
        dec_noisy = dec_measured + dec_delta

        # Sun angle stays the same (we want to know if it's *actually*
        # too bright to observe)

        # Collect actual observables (what you DO see)
        obs_measured: np.ndarray[Observation, np.dtype[np.float64]] = np.array(
            [
                Observation(
                    mjd=times[n],
                    ra=ra_noisy[n],
                    dec=dec_noisy[n],
                    sun_separation=sun_angle_deg[n],
                    sensor_id=self.id,
                )
                for n in range(num_obs)
            ]
        )

        # Remove observations that are not within the sensor limits

        ## Any viewing vector that is obscured by Earth is not valid
        sensor_dist = np.linalg.norm(x_sensor_truth, axis=1)
        viewing_vector = x_target - x_sensor_truth
        ### Calculate the angle from center of Earth, to sensor, to target.
        ### Note: If this angle is obtuse, it is automatically valid.
        viewing_angle = np.pi - np.abs(
            calc_separation_angle(x_sensor_truth, viewing_vector)
        )
        ### Calculate minimum height above Earth for viewing vector
        min_view_height = sensor_dist * np.sin(viewing_angle)
        view_valid = (min_view_height > R_earth) | (viewing_angle > np.pi / 2)

        ## Check the other sensor limits, then combine
        valid = view_valid * np.array([self.validate_limits(x) for x in obs_truth])

        obs_measured = obs_measured[valid]
        obs_truth = obs_truth[valid]
        obs_expected = obs_expected[valid]

        observations = ObservationCollection(
            pos_observed=obs_measured,
            pos_truth=obs_truth,
            pos_expected=obs_expected,
        )

        return observations


def pos_to_lat_lon(
    pos: NDArray[np.float64], times: NDArray[np.float64]
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Compute latitude & longitude for given position(s) and time(s).

    Parameters
    ----------
    pos : NDArray[np.float64]
        position vector
    times : NDArray[np.float64]
        array of timestamps

    Returns
    -------
    tuple[NDArray[np.float64], NDArray[np.float64]]
        computed latitudes and longitudes

    Raises
    ------
    ValueError
        The number of positions and times must be equal (for now)
    """

    # TODO(jake): refactor this code so as not to make the assumption that pos is a 2D array
    if len(pos) != len(times):
        raise ValueError("The number of positions and times must be equal (for now)")
    N = len(pos)

    # first rotate from TETED to ITRS
    pos_itrs = np.vstack([afc.TETEDToITRS(times[n], pos[n]) for n in range(N)])

    # compute lat/lon/alt from ITRS position
    lla = np.vstack(
        [afc.ITRSToLatLonAlt(pos_itrs[n]) for n in range(N)], dtype=np.float64
    )

    return lla[:, 0], lla[:, 1]


def spherical_to_cartesian(ra: float, dec: float) -> NDArray[np.float64]:
    """Convert spherical coordinates (ra, dec) to cartesian coordinates.

    Parameters
    ----------
    ra : float
        Right Ascension
    dec : float
        Declination

    Returns
    -------
    NDArray[np.float64]
        Cartesian coordinates
    """
    ra *= np.pi / 180
    dec *= np.pi / 180
    return np.array([np.cos(dec) * np.cos(ra), np.cos(dec) * np.sin(ra), np.sin(dec)])



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from typing import Callable, Tuple

import numpy as np
from numpy.typing import NDArray


class ImpulsiveManeuver:
    """
    ImpulseManeuver class holds an impulse maneuver definition, which includes the time of the maneuver, and the impulsive delta-v.

    Properties
    ----------
    time : float
        Timestamp of the maneuver (MJD, UTC)
    dv : NDArray[np.float64]
        3D Array of the impulsive delta-v (RSW frame)
    """

    _time: float
    _dv: NDArray[np.float64]

    def __init__(self, time: float, dv: NDArray[np.float64]):
        """Initialize the ImpulseManeuver class.

        Parameters
        ----------
        time : float
            Timestamp of the maneuver (MJD, UTC)
        dv : NDArray[np.float64]
            3D Array of the impulsive delta-v (RSW frame)
        """
        self._time = time
        self._dv = dv

    @property
    def time(self) -> float:
        return self._time

    @property
    def dv(self) -> NDArray[np.float64]:
        return self._dv


class ContinuousManeuver:
    """
    ContinuousManeuver class holds a continuous maneuver definition, which includes the acceleration
    function defining the maneuver, and the time range over which the maneuver occurs.
    """

    _accel_func: Callable
    _time_range: Tuple[float, float]

    def __init__(self, f: Callable, time_range: Tuple[float, float]):
        """Create a ContinuousManeuver object

        Parameters
        ----------
        f : Callable
            Function which evaluates the acceleration at a given time
        time_range : Tuple[float, float]
            Time range within which this maneuver is valid
        """
        self._accel_func = f
        if time_range[0] >= time_range[1]:
            a, b = time_range
            time_range = (b, a)

        self._time_range = time_range

    def __call__(self, t: float) -> NDArray[np.float64]:
        """Compute the acceleration for this maneuver at the time given

        Parameters
        ----------
        t : float
            Timestamp (MJD, UT1) to evaluate the acceleration

        Returns
        -------
        NDArray[np.float64]
            Acceleration vector, shape: (3,). (TETED, km/s^2)
        """
        if self._time_range[0] <= t <= self._time_range[1]:
            return self._accel_func(t)

        return np.zeros(3)



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import List, Sequence

import numpy as np
import yaml
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from numpy.typing import NDArray

from madlib._utils import MadlibException, sensor_yaml_schema

from ._observation import ObservationCollection, combineObsCollections
from ._satellite import Satellite
from ._sensor import GroundOpticalSensor, _Sensor


class SensorException(Exception):
    """SensorException class"""

    def __init__(self, message=None):
        super().__init__(message)


class SensorCollection:
    """Class containing multiple sensor objects that can generate a
    comprehensive observing schedule and collate observations."""

    def __init__(self, sensorList: Sequence[_Sensor]):
        """Initialize SensorCollection class

        Parameters
        ----------
        sensorList : List[Sensor]
            List of Sensors to include in observation network
        """
        self.sensorList = list(sensorList)
        self.numSensors = len(sensorList)

        self.obsTimes: List[NDArray[np.float64]] | None = None

    @staticmethod
    def paramsFromYAML(
        yaml_file: str | Path,
    ):
        """Parse a YAML into sensor parameters, returned as a list of dicts"""
        try:
            with open(yaml_file, "r") as f:
                sensor_data = yaml.safe_load(f)
                validate(sensor_data, sensor_yaml_schema)
        except ValidationError as e:
            if e.message == "'sensor_list' is a required property":
                msg = f"The sensor YAML file {yaml_file} must contain the top-level property 'sensor_list'."
            else:
                e_path = str(e.path)
                section = "/".join(list(e_path))
                msg = (
                    f"Error in the sensor YAML file [{yaml_file}] in the entry [{section}]: "
                    f"{e.message}"
                )
            raise MadlibException(msg) from None

        sensor_data = sensor_data["sensor_list"]

        return sensor_data

    @classmethod
    def fromYAML(cls, yaml_file: str):
        """Instantiate a SensorCollection object from a YAML file

        Parameters
        ----------
        yaml_file : str
            Path to YAML file defining the sensors in the collection
        """
        sensor_data = cls.paramsFromYAML(yaml_file)
        sensors = [GroundOpticalSensor(**params) for key, params in sensor_data.items()]
        sensor_network = cls(sensors)

        return sensor_network

    def generate_obs_timing(self, start: float, end: float):
        """Given a start time and an end time (in MJD), generate an
        array of observation times (also in MJD) based on the sensors'
        defined parameters.

        Parameters
        ----------
        start : float
            Earliest possible observation timestamp (MJD)
        end : float
            Latest possible observation (MJD)

        Raises
        ------
        SensorException
            The observation schedule has already been generated.
        """
        if self.obsTimes is not None:
            message = "The observation schedule has already been generated."
            raise SensorException(message=message)

        self.obsTimes = [
            sensor.generate_obs_timing(start, end) for sensor in self.sensorList
        ]

    def add_sensor(self, sensor: _Sensor):
        """Add a new sensor to the existing collection, provided the
        sensor timing has not already been generated.

        Parameters
        ----------
        sensor : _Sensor
            The sensor object to add to the collection
        """
        if self.obsTimes is not None:
            message = (
                "Cannot add new sensors to a SensorCollection "
                "if observation timing has already been generated."
            )
            raise (SensorException(message=message))

        self.sensorList.append(sensor)
        self.numSensors = len(self.sensorList)

    def observe(self, target_satellite: Satellite) -> ObservationCollection:
        """Given a madlib.Satellite, generate an ObservationCollection

        Parameters
        ----------
        target_satellite : Satellite
            madlib.Satellite is a class for propagating a satellite

        Returns
        -------
        ObservationCollection
            Observations of a satellite from multiple sensors, combined into a single object.

        Raises
        ------
        SensorException
            The observation schedule has already been generated.
        """
        if self.obsTimes is None:
            message = "obsTimes is None. Did you forget to generate obs timing first?"
            raise SensorException(message=message)
        else:
            obsCollections: List[ObservationCollection] = [
                sensor.observe(target_satellite, obstimes)
                for sensor, obstimes in zip(self.sensorList, self.obsTimes)
            ]
            observations: ObservationCollection = combineObsCollections(obsCollections)

        return observations



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import json
import shutil
from pathlib import Path
from typing import Callable, Tuple

import numpy as np
import pandas as pd
from hydra_zen import launch, make_config

from madlib import SensorCollection
from madlib._utils import MadlibException


class NotImplementedError(Exception):
    """NotImplementedError Exception"""

    pass


def create_task_fn(method: Callable) -> Callable:
    """Create a task function for hydra. Config (cfg) will be passed in via hydra.

    Parameters
    ----------
    method : Callable
        _description_

    Returns
    -------
    Callable
        _description_
    """

    def task_fn(cfg):
        try:
            output = method(**cfg)

            if output is not None:
                output.to_csv("output.csv", index=False)

        except MadlibException as e:
            with open("error.txt", "w") as f:
                f.write(str(e))

    return task_fn


def launcher(
    simulator_method: Callable,
    mtype: str,
    num_sim_pairs: int,
    sensor_yaml: str | Path,
    outdir: str | Path,
    dv_ric_mean_kms: Tuple[float, float, float] = (0.0, 0.0, 0.0),
    dv_ric_std_kms: Tuple[float, float, float] = (0.0, 0.1, 1.0),
    cont_thrust_duration_days: float | None = None,
    cont_thrust_mag: float = 1e-7,
    cont_thrust_model: int = 0,
    submitit: str = "",
    multirun_root: str = "",
    rm_multirun_root: bool = False,
    start_mjd: float | None = None,
    sim_duration_days: float = 3.0,
    random_seed: int | None = None,
    pred_err: float = 0.0,
    sensor_dra: float | None = None,
    sensor_ddec: float | None = None,
    sims_per_task: int = 1,
) -> None:
    """Hydra job launcher wrapper.

    Parameters
    ----------
    simulator_method : Callable
        Task function
    mtype : str
        Maneuver type:
        "impulse" = ImpulseManeuver,
        "continuous" = ContinuousManeuver,
    num_sim_pairs : int
        Number of simulations to perform per maneuver type.
    sensor_yaml : str | Path
        Path to YAML file defining the sensor network for the simulation
    outdir : str | Path
        Path to output directory where the concatenated results will be saved in a .csv file (complete.csv)
    dv_ric_mean_kms : Tuple[float, float, float], optional
        Mean values of normal distributions to use when sampling
        the radial, in-track, and cross-track delta-V values, respectively,
        of impulsive maneuvers. In units of km/s, by default [0.0, 0.0, 0.0]
    dv_ric_std_kms : Tuple[float, float, float], optional
        Standard deviations of normal distributions to use when sampling
        the radial, in-track, and cross-track delta-V values, respectively,
        of impulsive maneuvers. In units of km/s, by default [0.0, 0.1, 1.0]
    cont_thrust_duration_days : float | None, optional
        Duration in days of the continuous maneuver that begins at simulation start, by default None
        (if None, maneuver duration is equal to simulation duration)
    cont_thrust_mag : float, optional
        Magnitude of the continuous thrust (km/s/s), by default 1e-7
    cont_thrust_model : int, optional
        Which continuous thrust model to use:
        0 = applies a continuous thrust in the [0,1,0] direction,
        1 = applies a continuous thrust in a random direction,
        by default 0
    submitit : string, optional
        If specified, the path to a config JSON file defining how to launch jobs across multiple GPUs using submitit, by default None (serial launch only)
    multirun_root : string, optional
        If specified, the path to a directory where multirun output will be stored, by default None (./multirun will be used)
    rm_multirun_root : bool, optional
        Whether or not to delete the hydra multirun directory after finishing the simulation, by default False
    start_mjd : float, optional
        MJD at which the simulation should begin, by default None (current MJD)
    sim_duration_days : float, optional
        Duration of the simulation (days), by default 3.0
    random_seed : int, optional
        Random seed to use for numpy, by default None
    pred_err : float
        Fractional error on predicted initial orbital state
    sensor_dra : float, optional
        Sensor metric accuracy in the right ascension direction (arcsec).
        If not set, value is None, and `dra` value in sensor_yaml file
        will be used, by default: None
    sensor_ddec : float, optional
        Sensor metric accuracy in the declination direction (arcsec).
        If not set, value is None, and `dra` value in sensor_yaml file
        will be used, by default: None
    sims_per_task : int, optional
        Number of simulations to perform per task function, by default 1

    Raises
    ------
    NotImplementedError
        If an mtype was requested that is not yet implemented.
    """

    error_runs = []
    log_runs = []

    # Parse the sensor YAML file
    sensor_data = SensorCollection.paramsFromYAML(sensor_yaml)

    # Update sensor `dra` if was given as an input argument
    if sensor_dra is not None:
        for key in sensor_data.keys():
            sensor_data[key]["dra"] = sensor_dra

    # Update sensor `ddec` if was given as an input argument
    if sensor_ddec is not None:
        for key in sensor_data.keys():
            sensor_data[key]["ddec"] = sensor_ddec

    if cont_thrust_duration_days is None:
        cont_thrust_duration_days = sim_duration_days

    Conf = make_config(
        seq_id=0,
        sensor_params=sensor_data,
        maneuver_type=0,
        num_sim_pairs=num_sim_pairs,
        dv_ric_mean_kms=dv_ric_mean_kms,
        dv_ric_std_kms=dv_ric_std_kms,
        cont_thrust_duration_days=0,
        cont_thrust_mag=1e-7,
        cont_thrust_model=0,
        start_mjd=start_mjd,
        sim_duration_days=sim_duration_days,
        random_seed=random_seed,
        pred_err=pred_err,
        sims_per_task=sims_per_task,
    )

    # useful into to stout
    print(f"INFO :: {mtype = }")
    print(f"INFO :: {sim_duration_days = }")
    print(f"INFO :: {sims_per_task = }")

    # hydra_zen overrides: initialize
    overrides = []

    # hydra_zen overrides: mtype specific cases
    if mtype == "impulse":
        overrides += [
            f"maneuver_type=0,1",
        ]
    elif mtype == "continuous":
        print(f"INFO :: {cont_thrust_duration_days = }")
        print(f"INFO :: {cont_thrust_mag = }")
        print(f"INFO :: {cont_thrust_model = }")
        overrides += [
            f"maneuver_type=0,2",
            f"cont_thrust_duration_days={cont_thrust_duration_days}",
            f"cont_thrust_mag={cont_thrust_mag}",
            f"cont_thrust_model={cont_thrust_model}",
        ]
    else:
        # all mtypes case?
        # maneuver_type=0,1,2
        raise NotImplementedError("An mtype was requested that is not yet implemented.")

    # create task function
    task_fn = create_task_fn(simulator_method)

    # define seq_id string
    # seq_id = ",".join([str(n) for n in range(num_sim_pairs)])
    seq_id = ",".join([str(n) for n in np.arange(0, num_sim_pairs, sims_per_task)])

    # setup hydra_zen overrides
    overrides += [
        f"seq_id={seq_id}",
    ]

    # configure submitit if using
    if submitit != "":
        with open(submitit, "r") as f:
            submitit_overrides = json.load(f)

        # Make sure the JSON contents are a list of strings
        str_list_check = all(isinstance(n, str) for n in submitit_overrides)
        if (type(submitit_overrides) != list) or (not str_list_check):
            raise MadlibException(
                f"The specified submitit configuration file {submitit} is not formatted properly. "
                "The JSON must be a list of strings."
            )

        overrides += submitit_overrides

    else:
        overrides += [
            "hydra.job.chdir=True",
        ]

    # configure multirun directory if non-default
    if multirun_root != "":
        # Append datetime structure to root
        multirun_dir = Path(multirun_root) / "${now:%Y-%m-%d}/${now:%H-%M-%S}"
        multirun_dir = str(multirun_dir)
        overrides += [
            f"hydra.sweep.dir={multirun_dir}",
        ]

    # setup outdir
    outdir = Path(outdir)
    outdir.mkdir(parents=True, exist_ok=True)

    # launch jobs
    (jobs,) = launch(
        Conf,
        task_fn,
        multirun=True,
        to_dictconfig=True,
        overrides=overrides,
        version_base="1.3",
    )

    # initialize df_merged dataframe that will contain all output.csv files concatenated
    df_merged = pd.DataFrame()

    # get multirun root directory
    rundir = Path(jobs[0].working_dir).parent

    # concatenate all output.csv files
    jobfiles = sorted(rundir.rglob("output.csv"))
    for csv_file in jobfiles:
        df_temp = pd.read_csv(csv_file)
        df_merged = pd.concat((df_merged, df_temp), ignore_index=True)

    # export merged to disk
    df_merged.to_csv(outdir / "complete.csv", index=False)

    # concatenate all error.txt files
    errfiles = sorted(rundir.rglob("error.txt"))
    for err in errfiles:
        with open(err, "r") as f:
            text = f.read()
        error_runs.append((err, text))

    # concatenate all zen_launch.log files
    logfiles = sorted(rundir.rglob("zen_launch.log"))
    for logfile in logfiles:
        with open(logfile, "r") as f:
            text = f.read()
        log_runs.append((logfile, text))

    # keep track of errors checkpoint-style
    with open(outdir / "errors.txt", "w") as f:
        for errfile, errtext in error_runs:
            f.write(str(errfile) + "\n")
            f.write(errtext + "\n\n")

    # keep track of zen_launch logs checkpoint-style
    with open(outdir / "logs.txt", "w") as f:
        for logfile, logtext in log_runs:
            if logtext != "":
                f.write(str(logfile) + "\n")
                f.write(logtext + "\n\n")

    # export merged to disk
    df_merged.to_csv(outdir / "complete.csv", index=False)

    # copy multirun.yaml file to outdir if exists
    try:
        multirun_yaml = next(rundir.glob("multirun.yaml"))
        shutil.copy2(multirun_yaml, outdir / "multirun.yaml")
    except (StopIteration, OSError):
        pass

    # remove multirun root directory?
    if rm_multirun_root:
        # DIRECTORIES WILL ONLY BE REMOVED IF THEIR CONTENTS ARE
        # EXCLUSIVELY HYDRA-ZEN MULTIRUN OUTPUTS
        for job in jobs:
            working_dir = Path(job.working_dir)
            hydra_dir = working_dir / ".hydra"
            if hydra_dir.exists() and hydra_dir.is_dir():
                shutil.rmtree(hydra_dir)

            (working_dir / "zen_launch.log").unlink(missing_ok=True)
            (working_dir / "output.csv").unlink(missing_ok=True)
            (working_dir / "error.txt").unlink(missing_ok=True)

            contents = list(working_dir.glob("*"))
            if len(contents) == 0:
                shutil.rmtree(working_dir)

        if (rundir / ".submitit").exists():
            shutil.rmtree(rundir / ".submitit")

        parent = rundir.parent
        grandparent = parent.parent

        (rundir / "multirun.yaml").unlink(missing_ok=True)
        contents = list(rundir.glob("*"))
        if len(contents) == 0:
            shutil.rmtree(rundir)

        parent_contents = list(parent.glob("*"))
        if len(parent_contents) == 0:
            shutil.rmtree(parent)

        grandparent_contents = list(grandparent.glob("*"))
        if len(grandparent_contents) == 0:
            shutil.rmtree(grandparent)



# Copyright (c) 2024 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

"""
Test file: test_madlib_satellite.py
Description: This file contains unit tests which test "edge cases" in
    the madlib._satellite module.
"""

import pathlib
import sys

import astroforge as af
import numpy as np
from astropy.time import Time

# add parent directory of __file__ to sys.path, if isn't already included
if str(pathlib.Path(__file__).parents[1]) not in sys.path:
    sys.path.append(str(pathlib.Path(__file__).parents[1]))

import madlib
from madlib._utils import MadlibException
from madlib._satellite import Satellite, ContinuousThrustSatellite
from madlib._maneuver import ImpulsiveManeuver

import pytest


seed = 4445


def test_Satellite_kwargs_NotImplementedError():
    _epoch = 54321.0
    _pos = np.zeros((3,))
    _vel = np.zeros((3,))
    _accel = np.zeros((3,))
    _manuever_info = None
    with pytest.raises(NotImplementedError):
        Satellite(
            _epoch,
            _pos,
            _vel,
            _accel,
            _manuever_info,
            today="Tuesday",  # **kwargs, should raise NotImplementedError
        )


def test_Satellite_property_and_setters():
    _epoch = 54321.0
    _pos = np.zeros((3,))
    _vel = np.zeros((3,))
    _accel = np.zeros((3,))

    # --- Satellite w/ ImpulseManeuver
    _manuever_info = None
    sat = Satellite(
        _epoch,
        _pos,
        _vel,
        _accel,
        _manuever_info,
    )

    maneuver = ImpulsiveManeuver(
        54321.0 + 1.0,  # _time (mjd)
        np.ones((3,)),  # _dv
    )

    # --- @maneuver.setter
    assert sat.maneuver == None
    sat.maneuver = maneuver
    assert sat.maneuver == maneuver

    # --- @property : epoch
    assert sat.epoch == _epoch

    # --- @epoch.setter & @property
    assert sat.epoch == _epoch
    sat.epoch = _epoch + 1
    assert sat.epoch == _epoch + 1

    # --- @x.setter & @property
    assert all(sat.x == _pos)
    sat.x = _pos + 1
    assert all(sat.x == _pos + 1)

    # --- @v.setter & @property
    assert all(sat.v == _vel)
    sat.v = _vel + 1
    assert all(sat.v == _vel + 1)

    # --- @a.setter & @property
    assert all(sat.acc == _accel)
    sat.acc = _accel + 1
    assert all(sat.acc == _accel + 1)

    # --- Satellite.copy() test
    sat_copy = sat.copy()
    assert all(
        [
            sat_copy.does_maneuver == sat.does_maneuver,
            sat_copy.maneuver.time == sat.maneuver.time,  # type: ignore
            all(sat_copy.maneuver.dv == sat.maneuver.dv),  # type: ignore
            sat_copy.epoch == sat.epoch,
            all(sat_copy.x == sat.x),
            all(sat_copy.v == sat.v),
            all(sat_copy.acc == sat.acc),
        ]
    )

    # --- __str__() and __repr__() methods test
    sat_str = str(sat)
    sat_repr = repr(sat)
    assert isinstance(sat_str, str)
    assert isinstance(sat_repr, str)


def test_Satellite_invalid_instantiations():
    _epoch = 54321.0
    _pos = [1, 2, 3]
    _vel = np.zeros((3,))
    _accel = np.zeros((3,))

    with pytest.raises(TypeError):
        Satellite(
            _epoch,
            _pos,  # type: ignore
            _vel,
            _accel,
        )

    _pos = np.zeros((1, 3))

    with pytest.raises(ValueError):
        Satellite(
            _epoch,
            _pos,
            _vel,
            _accel,
        )


def test_Satellite_propagate_method(seed=seed):
    np.random.seed(seed)
    # --- orbital state
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- maneuver definition
    man_time = epoch + 5.0 / 24
    man_dv = np.array([0.0, 5.0, 0.0]) / 1000
    man = madlib.ImpulsiveManeuver(man_time, man_dv)

    # --- define satellites
    sat0 = madlib.Satellite(epoch, x0, v0, a0)  # maneuver = False
    sat1 = madlib.Satellite(epoch, x0, v0, a0, man)  # manuever = True

    # --- propagate, times is a scalar
    _times = epoch + 1
    _X, _V = sat0.propagate(_times)
    assert np.allclose(_X, [2.83737953e04, 3.11888127e04, -3.64526039e-01])
    assert np.allclose(_V, [-2.27434029e00, 2.06902511e00, 1.99407148e-04])

    # --- propagate, raise ValueError: One of the propagation times is the exact same time as the maneuver.
    with pytest.raises(ValueError):
        sat1.propagate(sat1.maneuver.time)  # type: ignore


def test_Satellite_from_GEO_longitude(seed=seed):
    np.random.seed(seed)
    # Satellite
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd
    sat_longitude = 360 * np.random.random()
    sat_observed = madlib.Satellite.from_GEO_longitude(sat_longitude, epoch)

    assert sat_observed.epoch == epoch
    assert isinstance(sat_observed, madlib._satellite.Satellite)


def test_Satellite_from_keplerian(seed=seed, rtol=1e-3, atol=1e-3):
    np.random.seed(seed)
    # First, let's create a satellite from Cartesian coordinates we can use for comparison
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd
    sat0 = madlib.Satellite(epoch, x0, v0, a0)
    assert all(sat0.x == x0)
    assert all(sat0.v == v0)
    assert all(sat0.acc == a0)

    # Convert Cartesian -> Keplerian elements
    keplerian_elements = af.coordinates.cartesian_to_keplerian(sat0.x, sat0.v)

    # Create Satellite from Keplerian elements
    sat1 = madlib.Satellite.from_keplerian(
        epoch=epoch,
        inclination_rad=keplerian_elements["inclination_rad"],
        raan_rad=keplerian_elements["raan_rad"],
        argp_rad=keplerian_elements["argp_rad"],
        ecc=keplerian_elements["eccentricity"],
        semi_major_axis_km=keplerian_elements["semi_major_axis_km"],
        mean_anomaly_rad=keplerian_elements["mean_anomaly_rad"],
    )

    # Check: sat1 should be the same as sat0
    assert np.allclose(sat0.x, sat1.x, rtol=rtol, atol=atol)
    assert np.allclose(sat0.v, sat1.v, rtol=rtol, atol=atol)
    assert np.allclose(sat0.acc, sat1.acc, rtol=rtol, atol=atol)

    # Let's do it again but with a bunch of randomly created satellite orbits
    for i in range(1000):
        sat_longitude = 360 * np.random.random()
        sat_rand = madlib.Satellite.from_GEO_longitude(sat_longitude, epoch)
        keplerian_elements = af.coordinates.cartesian_to_keplerian(
            sat_rand.x, sat_rand.v
        )
        sat_from_keplerian = madlib.Satellite.from_keplerian(
            epoch=epoch,
            inclination_rad=keplerian_elements["inclination_rad"],
            raan_rad=keplerian_elements["raan_rad"],
            argp_rad=keplerian_elements["argp_rad"],
            ecc=keplerian_elements["eccentricity"],
            semi_major_axis_km=keplerian_elements["semi_major_axis_km"],
            mean_anomaly_rad=keplerian_elements["mean_anomaly_rad"],
        )
        assert np.allclose(sat_rand.x, sat_from_keplerian.x, rtol=rtol, atol=atol)
        assert np.allclose(sat_rand.v, sat_from_keplerian.v, rtol=rtol, atol=atol)
        assert np.allclose(sat_rand.acc, sat_from_keplerian.acc, rtol=rtol, atol=atol)


def test_ContinuousThrustSatellite_edge_cases(seed=seed):
    np.random.seed(seed)
    # --- orbital state
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- ContinuousThrustSatellite w/ ContinuousManeuver
    # ContinuousThrust Function Definition
    def acc_f(t):
        return np.array([0.0, 1.0e-7, 0.0])

    acc_t_range = (epoch, epoch + 3)
    # ContinuousManeuver Definition
    man = madlib.ContinuousManeuver(acc_f, acc_t_range)
    sat = madlib.ContinuousThrustSatellite(
        epoch,
        x0,
        v0,
        a0,
        man,
    )

    # --- check case: _time_range input to ContinuousManeuver is reversed
    acc_t_range_rev = (acc_t_range[1], acc_t_range[0])
    man_rev_t = madlib.ContinuousManeuver(acc_f, acc_t_range_rev)
    assert acc_t_range == man_rev_t._time_range

    # --- propagate, times is a scalar
    _times = epoch + 1
    _X, _V = sat.propagate(_times)
    assert all(
        np.isclose(_X, np.array([2.79082377e04, 3.14487181e04, -3.31180976e-01]))
    )
    assert all(
        np.isclose(_V, np.array([-2.29974384e00, 2.05373079e00, 2.00473790e-04]))
    )


def test_ContinuousThrustSatellite_invalid_propagate(seed=seed):
    """Trying to apply a continuous thrust on an ordinary satellite should fail."""
    np.random.seed(seed)
    # --- orbital state
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # --- ContinuousThrustSatellite w/ ContinuousManeuver
    # ContinuousThrust Function Definition
    def acc_f(t):
        return np.array([0.0, 1.0e-7, 0.0])

    acc_t_range = (epoch, epoch + 3)
    # ContinuousManeuver Definition
    man = madlib.ContinuousManeuver(acc_f, acc_t_range)
    sat = madlib.Satellite(
        epoch,
        x0,
        v0,
        a0,
        man,
    )

    failed = False
    try:
        sat.propagate(epoch + 1)
    except MadlibException:
        failed = True

    assert failed


def test_ContinuousThrustSatellite_no_maneuver(seed=seed):
    """A ContinuousThrustSatellite with no maneuver should behave like
    a regular Satellite with no maneuver."""
    np.random.seed(seed)
    # --- orbital state
    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd

    # ContinuousManeuver Definition
    sat_c = madlib.ContinuousThrustSatellite(epoch, x0, v0, a0, None)

    # Ordinary satellite with no maneuver
    sat_n = madlib.Satellite(epoch, x0, v0)

    _times = epoch + 1
    Xc, Vc = sat_c.propagate(_times)
    Xn, Vn = sat_n.propagate(_times)
    assert np.allclose(Xc, Xn)
    assert np.allclose(Vc, Vn)


class TestTrueOrbits:

    th = 2 * np.pi * np.random.rand()
    x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
    v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
    a0 = np.zeros((3,))
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch: float = t_start.mjd

    th_true = th + 0.1
    x0_true = af.constants.Rgeo * np.array([np.cos(th_true), np.sin(th_true), 0.001])
    v0_true = af.constants.Vgeo * np.array([-np.sin(th_true), np.cos(th_true), 0.001])
    a0_true = a0 + 1e-5
    epoch_true = epoch + 0.1

    sat_default = madlib.Satellite(epoch, x0, v0)

    sat_modified = madlib.Satellite(
        epoch,
        x0,
        v0,
        a0,
        epoch_true=epoch_true,
        pos_true=x0_true,
        vel_true=v0_true,
        acc_true=a0_true,
    )

    def test_define_true_orbit(self):

        assert self.sat_default.epoch == self.sat_default.epoch_true
        assert self.sat_default.x_true is not None
        assert all(np.isclose(self.sat_default.x, self.sat_default.x_true))
        assert all(np.isclose(self.sat_default.v, self.sat_default.v_true))
        assert all(np.isclose(self.sat_default.acc, self.sat_default.acc_true))

        assert self.sat_modified.epoch != self.sat_modified.epoch_true
        assert self.sat_modified.x_true is not None
        assert all(~np.isclose(self.sat_modified.x, self.sat_modified.x_true))
        assert all(~np.isclose(self.sat_modified.v, self.sat_modified.v_true))
        assert all(~np.isclose(self.sat_modified.acc, self.sat_modified.acc_true))

    def test_propagate_true_orbit(self):
        x_default, v_default = self.sat_default.propagate(self.epoch + 0.5)
        x_default_true, v_default_true = self.sat_default.propagate(
            self.epoch + 0.5, use_true_orbit=True
        )

        x_modified, v_modified = self.sat_modified.propagate(self.epoch + 0.5)
        x_modified_true, v_modified_true = self.sat_modified.propagate(
            self.epoch + 0.5, use_true_orbit=True
        )

        assert np.allclose(x_default, x_default_true)
        assert np.allclose(v_default, v_default_true)
        assert ~np.allclose(x_modified, x_modified_true)
        assert ~np.allclose(v_modified, v_modified_true)

    def test_propagate_true_orbit_continuous(self):
        np.random.seed(0)
        # --- orbital state
        th = 2 * np.pi * np.random.rand()
        x0 = af.constants.Rgeo * np.array([np.cos(th), np.sin(th), 0.0])
        v0 = af.constants.Vgeo * np.array([-np.sin(th), np.cos(th), 0.0])
        a0 = np.zeros((3,))
        t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
        epoch = t_start.mjd

        # --- ContinuousThrustSatellite w/ ContinuousManeuver
        # ContinuousThrust Function Definition
        def acc_f(t):
            return np.array([0.0, 1.0e-7, 0.0])

        acc_t_range = (epoch, epoch + 3)
        # ContinuousManeuver Definition
        man = madlib.ContinuousManeuver(acc_f, acc_t_range)

        th_true = th + 0.1
        x0_true = af.constants.Rgeo * np.array(
            [np.cos(th_true), np.sin(th_true), 0.001]
        )
        v0_true = af.constants.Vgeo * np.array(
            [-np.sin(th_true), np.cos(th_true), 0.001]
        )
        a0_true = a0 + 1e-5
        epoch_true = epoch + 0.1

        sat = madlib.ContinuousThrustSatellite(
            epoch,
            x0,
            v0,
            a0,
            man,
            epoch_true=epoch_true,
            pos_true=x0_true,
            vel_true=v0_true,
            acc_true=a0_true,
        )

        assert sat.epoch != sat.epoch_true
        assert sat.x_true is not None
        assert ~np.allclose(sat.x, sat.x_true)
        assert ~np.allclose(sat.v, sat.v_true)
        assert ~np.allclose(sat.acc, sat.acc_true)

        x, v = sat.propagate(epoch + 1.0)
        x_true, v_true = sat.propagate(epoch + 1.0, use_true_orbit=True)

        assert ~np.allclose(x, x_true)
        assert ~np.allclose(v, v_true)

    def test_setters(self):
        self.sat_default.epoch_true = self.epoch_true
        self.sat_default.x_true = self.x0_true
        self.sat_default.v_true = self.v0_true
        self.sat_default.acc_true = self.a0_true

        assert self.sat_default.epoch != self.sat_default.epoch_true
        assert self.sat_default.x_true is not None
        assert ~np.allclose(self.sat_default.x, self.sat_default.x_true)
        assert ~np.allclose(self.sat_default.v, self.sat_default.v_true)
        assert ~np.allclose(self.sat_default.acc, self.sat_default.acc_true)


class TestCrossTags:
    t_start = Time("2011-11-11T11:11:11", format="isot", scale="utc")
    epoch = t_start.mjd
    sat = madlib.Satellite.from_GEO_longitude(lon=0.0, epoch=epoch)

    def test_zero_cross_tag(self):
        cross_sat = self.sat.create_cross_tag(
            cross_mjd=self.epoch,
            delta_pos_km=np.zeros(3),
            delta_vel_kms=np.zeros(3),
        )

        assert np.isclose(cross_sat.epoch, self.sat.epoch, atol=0.1, rtol=0)
        assert np.allclose(cross_sat.x, self.sat.x)
        assert np.allclose(cross_sat.v, self.sat.v)
        assert np.allclose(cross_sat.acc, self.sat.acc)

        x, v = self.sat.propagate(self.epoch + 1)
        x_cross, v_cross = cross_sat.propagate(self.epoch + 1)

        assert np.allclose(x, x_cross)
        assert np.allclose(v, v_cross)

    def test_nonzero_cross_tag(self):
        cross_sat = self.sat.create_cross_tag(
            cross_mjd=self.epoch + 0.5,
            delta_pos_km=np.array([10, 10, 10]),
            delta_vel_kms=np.array([-5, -5, -5]),
        )

        assert ~np.isclose(cross_sat.epoch, self.sat.epoch, atol=0.1, rtol=0)
        assert ~np.allclose(cross_sat.x, self.sat.x)
        assert ~np.allclose(cross_sat.v, self.sat.v)

        x, v = self.sat.propagate(np.array([self.epoch + 0.5, self.epoch + 1.0]))
        x_cross, v_cross = cross_sat.propagate(
            np.array([self.epoch + 0.5, self.epoch + 1.0])
        )

        assert np.allclose(x[0] + 10, x_cross[0], rtol=0, atol=0.1)
        assert np.allclose(v[0] - 5, v_cross[0], rtol=0, atol=0.1)

        assert ~np.allclose(x[1], x_cross[1])
        assert ~np.allclose(v[1], v_cross[1])
